import pytest
import pandas as pd
from usdm_excel.base_sheet import BaseSheet
from usdm_model.code import Code

xfail = pytest.mark.xfail

def test_cell_empty(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': [None, 'b', 'c', None]}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,0,False),
    (3,0,False),
    (0,1,True),
    (3,1,True)
  ]
  for test in test_data:
    assert(base.cell_empty(test[0],test[1])) == test[2]

def test_cell_empty_legacy(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': [None, 'b', '-', None]}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,0,'3',False),
    (3,0,'0',False),
    (0,1,'',True),
    (2,1,'',True),
    (3,1,'',True)
  ]
  for test in test_data:
    value, is_null = base.read_cell_empty_legacy(test[0],test[1])
    assert(value) == test[2]
    assert(is_null) == test[3]

def test_read_cell_empty(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': [None, 'b', '-', None]}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,0,'3','3'),
    (3,0,'0','0'),
    (0,1,'',''),
    (2,1,'','-'),
    (3,1,'','')
  ]
  for test in test_data:
    assert(base.read_cell_empty(test[0],test[1],'-')) == test[2]
    assert(base.read_cell_empty(test[0],test[1],'=')) == test[3]

def test_read_cell(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', 'b', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,0,'3'),
    (3,0,'0'),
    (0,1,'a'),
    (3,1,'')
  ]
  for test in test_data:
    assert(base.read_cell(test[0],test[1])) == test[2]

def test_read_cell_default(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', 'b', 'c', None]}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (3,1,'100s','100s')
  ]
  for test in test_data:
    assert(base.read_cell(test[0],test[1], default=test[2])) == test[3]

def test_read_cell_error(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', 'b', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  assert(base.read_cell(6,8)) == ""
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "sheet"
  assert mock_error.call_args[0][1] == 7
  assert mock_error.call_args[0][2] == 9
  assert mock_error.call_args[0][3] == "Exception. Error reading cell. See log for additional details."
  
def test_read_cell_by_name(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['tom', 10], ['nick', 15], ['juli', 14], ['fred', '']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Age'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,'Name','tom'),
    (2,'Name','juli'),
    (0,'Age','10'),
    (2,'Age','14'),
    (3,'Age',''),
  ]
  for test in test_data:
    assert(base.read_cell_by_name(test[0],test[1])) == test[2]

def test_read_cell_by_name_default(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['tom', 10], ['nick', 15], ['juli', 14], ['fred', '']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Age'])
  base = BaseSheet("", globals, "sheet")
  assert(base.read_cell_by_name(0, 'NameX', 'Default')) == 'Default'
  assert(base.read_cell_by_name(0, 'Name', 'Default')) == 'tom'
  assert(base.read_cell_by_name(0, ['NameX', 'XXX'], 'Default')) == 'Default'
  assert(base.read_cell_by_name(0, ['NameX', 'XXX', 'Age'], 'Default')) == '10'
  assert(base.read_cell_by_name(3, 'Age', default='100')) == ''

def test_read_cell_by_name_error(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['tom', 10], ['nick', 15], ['juli', 14]]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Age'])
  base = BaseSheet("", globals, "sheet")
  assert(base.read_cell_by_name(1,'Not There')) == ""
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "sheet"
  assert mock_error.call_args[0][1] == 2
  assert mock_error.call_args[0][2] == -1
  assert mock_error.call_args[0][3] == "Error attempting to read cell 'Not There'. Exception: Failed to detect column(s) 'Not There' in sheet"

def test_read_cell_by_name_present(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['tom', 10], ['nick', 15], ['juli', 14]]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Age'])
  base = BaseSheet("", globals, "sheet")
  assert(base.read_cell_by_name(1,'Not There',must_be_present=False)) == ""
  mock_error.assert_not_called()

def test_read_cell_multiple(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['tom', ''], 
    ['nick', 'Sam'], 
    ['juli', ' Fred, Dick,   Harry  '], 
    ['andy', 'John  , Jane'], 
    ['andy', '"John, & Fred", Jane'],
    ['andy', '"John, \\" & Fred", Jane']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,1,[]),
    (2,1,['Fred', 'Dick', 'Harry']),
    (3,1,['John', 'Jane']),
    (4,1,['John, & Fred', 'Jane']),
    (5,1,['John, " & Fred', 'Jane'])
  ]
  for test in test_data:
    assert(base.read_cell_multiple(test[0],test[1])) == test[2]

def test_read_cell_multiple_by_name(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['tom', ''], 
    ['nick', 'Sam'], 
    ['juli', ' Fred, Dick,   Harry  '], 
    ['andy', 'John  , Jane'], 
    ['andy', '"John, & Fred", Jane'],
    ['andy', '"John, \\" & Fred", Jane']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,[],                         True,  ""),
    (2,['Fred', 'Dick', 'Harry'],  True,  ""),
    (3,['John', 'Jane'],           True,  ""),
    (4,['John, & Fred', 'Jane'],   True,  ""),
    (5,['John, " & Fred', 'Jane'], True,  ""),
    (6,[],                         False, "")
  ]
  for test in test_data:
    errors = globals.errors_and_logging._errors
    errors.clear()
    assert(base.read_cell_multiple_by_name(test[0],'Children')) == test[1]
    
    #assert(errors.items[0].to_log()) == test[3]

# read_cell_with_previous

def test_read_boolean_cell_by_name(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  #data = [[0, 1, 2, 3, 4, 5, 6, 7, 8], ['a', 'y', 'Y', 'true', 'True', 'yes', 1, '1', '']]
  data = [[0, 'a'], [1, 'y'], [2, 'Y'], [3, 'true'], [4, 'True'], [5, 'yes'], [6, 1,], [7, '1'], [8, ''], [9, 'T']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0,'Children',False),
    (1,'Children',True),
    (2,'Children',True),
    (3,'Children',True),
    (4,'Children',True),
    (5,'Children',True),
    (6,'Children',True),
    (7,'Children',True),
    (8,'Children',False),
    (9,'Children',True),
  ]
  for test in test_data:
    assert(base.read_boolean_cell_by_name(test[0],test[1])) == test[2]

def test_read_quantity_cell_by_name(mocker, globals):
  globals.cross_references.clear()
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [[''], ['1 days'], [' 1 weeks'], ['14 Hours'], ['14'], ['']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Quantity'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    #  Name,       Allow Missing Units, Allow Empty, Errors, Empty,  Value,  Unit,     Error 
    (0,'Quantity', False,               False,       True,   False,  0.0,    '',       "Error in sheet sheet at [1,1]: Could not decode the quantity value, appears to be empty ''"),
    (1,'Quantity', False,               False,       False,  False,  1.0,    'C25301', ''),
    (2,'Quantity', False,               False,       False,  False,  1.0,    'C29844', ''),
    (3,'Quantity', False,               False,       False,  False,  14.0,   'C25529', ""),
    (4,'Quantity', True,                False,       False,  False,  14.0,   '',       ""),
    (5,'Quantity', False,               True,        False,  True,   0.0,    '',       ""),
  ]
  for test in test_data:
    errors = globals.errors_and_logging._errors
    errors.clear()
    item = base.read_quantity_cell_by_name(test[0],test[1],test[2],test[3]) 
    if not test[4] and not test[5]:
      assert(item.value) == test[6]
      if not test[2]:
        assert(item.unit.standardCode.code) == test[7]
      assert(len(errors.items)) == 0
    elif test[5]:
      assert(item) == None
    else:
      assert(item) == None
      assert(errors.items[0].to_log()) == test[8]

def test_read_range_cell_by_name(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [[''], ['1..2 days'], [' 1 .. 3 weeks'], [' .. 3 weeks'], ['1 ..  weeks'], ['1 . 4 weeks'], ['14 Hours'], ['14'], ['']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Range'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    #  Name,    Req Units, Allow Empty, Errors, Empty, Min,  Max,   Unit,     Error 
    (0,'Range', True,      False,       True,   False, 0.0,  0.0,   '',       "Error in sheet sheet at [1,1]: Could not decode the range value, appears to be empty ''"),
    (1,'Range', True,      False,       False,  False, 1.0,  2.0,   'C25301', ''),
    (2,'Range', True,      False,       False,  False, 1.0,  3.0,   'C29844', ''),
    (3,'Range', True,      False,       True,   False, 0.0,  0.0,   '',       "Error in sheet sheet at [4,1]: Could not decode the range value '.. 3 weeks'"),
    (4,'Range', True,      False,       True,   False, 0.0,  0.0,   '',       "Error in sheet sheet at [5,1]: Unable to set the units code for the range '1 ..  weeks'"),
    (5,'Range', True,      False,       True,   False, 0.0,  0.0,   '',       "Error in sheet sheet at [6,1]: Unable to set the units code for the range '1 . 4 weeks'"),
    (6,'Range', True,      False,       False,  False, 14.0, 14.0,  'C25529', ""),
    (7,'Range', False,     False,       False,  False, 14.0, 14.0,  '',       ""),
    (8,'Range', True,      True,        False,  True,  0.0,  0.0,   '',       ""),
  ]
  for test in test_data:
    globals.errors_and_logging.errors().clear()
    #print(f"INDEX: {test[0]}")
    range = base.read_range_cell_by_name(test[0],test[1],test[2],test[3]) 
    if not test[4] and not test[5]:
      assert(range.minValue) == test[6]
      assert(range.maxValue) == test[7]
      if test[2]:
        assert(range.unit.code) == test[8]
      assert(range.isApproximate) == False
      assert(len(globals.errors_and_logging.errors().items)) == 0
    elif test[5]:
      assert(range) == None
    else:
      assert(range) == None
      assert(globals.errors_and_logging.errors().items[0].to_log()) == test[9]

# def test_read_description_by_name(mocker, globals):
#   mock_option = mocker.patch("usdm_excel.om.get")
#   mock_option.side_effect=['xxx', 'xxx', 'xxx']
#   mocked_open = mocker.mock_open(read_data="File")
#   mocker.patch("builtins.open", mocked_open)
#   data = [[0, ''], [1, 'something'], [2, '  ']]
#   mock_read = mocker.patch("pandas.read_excel")
#   mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
#   base = BaseSheet("", globals, "sheet")
#   test_data = [
#     (0,'Children','xxx'),
#     (1,'Children','something'),
#     (2,'Children','xxx'),
#   ]
#   for test in test_data:
#     assert(base.read_description_by_name(test[0],test[1])) == test[2]

@xfail
def test_read_cell_with_previous():
  assert 0

@xfail
def test_read_other_code_cell_by_name():
  assert 0

@xfail
def test_read_other_code_cell():
  assert 0

@xfail
def test_read_other_code_cell_multiple_by_name():
  assert 0

@xfail
def test_read_other_code_cell_mutiple():
  assert 0

def test_column_present(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", globals, "sheet")
  assert(base.column_present("Name")) == 0
  assert(base.column_present(["Name"])) == 0
  assert(base.column_present(["Children"])) == 1
  assert(base.column_present(["ChildrenX", "Children"])) == 1
  with pytest.raises(BaseSheet.FormatError):
    assert(base.column_present(["Fred"]))
  
def test_read_cdisc_klass_attribute_cell_by_name(mocker, globals):
  expected = Code(id='CodeX', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected, None]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [[0, 'a'], [1, ''], [2, 'c']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0, 'Children', expected, ""),
    (1, 'Children', None, "sheet", 2, 2, "Empty cell detected where CDISC CT value expected."),
    (2, 'Children', None, "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell_by_name( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]

def test_read_cdisc_klass_attribute_cell(mocker, globals):
  expected = Code(id='CodeX', code='code', codeSystem='codesys', codeSystemVersion='3', decode="label")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected, None]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a', '', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0, 1, expected, ""),
    (1, 1, None, "sheet", 2, 2, "Empty cell detected where CDISC CT value expected."),
    (2, 1, None, "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]
     
@xfail
def test_read_cdisc_klass_attribute_cell_multiple_by_name(mocker, globals):
  assert 0
  
def test_read_cdisc_klass_attribute_cell_multiple(mocker, globals):
  expected_1 = Code(id='CodeX1', code='code1', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(id='CodeX2', code='code2', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, None]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'col_1': [3, 2, 1, 0], 'col_2': ['a,b', '', 'c', '']}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0, 1, [expected_1, expected_2], ""),
    (1, 1, [], "sheet", 2, 2, "Empty cell detected where multiple CDISC CT values expected."),
    (2, 1, [], "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell_multiple( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]

def test_read_cdisc_klass_attribute_cell_multiple_by_name(mocker, globals):
  expected_1 = Code(id='CodeX1', code='code1', codeSystem='codesys', codeSystemVersion='3', decode="label1")
  expected_2 = Code(id='CodeX2', code='code2', codeSystem='codesys', codeSystemVersion='3', decode="label2")
  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[expected_1, expected_2, None]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [[0, 'a,b'], [1, ''], [2, 'c']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['Name', 'Children'])
  base = BaseSheet("", globals, "sheet")
  test_data = [
    (0, 'Children', [expected_1, expected_2], ""),
    (1, 'Children', [], "sheet", 2, 2, "Empty cell detected where multiple CDISC CT values expected."),
    (2, 'Children', [], "sheet", 3, 2, "CDISC CT not found for value 'c'.")
  ]
  for test in test_data:
    assert(base.read_cdisc_klass_attribute_cell_multiple_by_name( 'X', 'y', test[0], test[1])) == test[2]
    if not test[3] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[3]
      assert mock_error.call_args[0][1] == test[4]
      assert mock_error.call_args[0][2] == test[5]
      assert mock_error.call_args[0][3] == test[6]

def test__decode_other_cell(mocker, globals):
  expected = Code(id='Code_1', code='c', codeSystem='a', codeSystemVersion='3', decode="d")
  mock_version = mocker.patch("usdm_excel.ct_version_manager.CTVersionManager.get")
  mock_version.side_effect=['3']
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1']
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    ("", 0, 0, None, ""),
    ("xxx", 1, 1, None, "sheet", 2, 2, "Failed to decode code data 'xxx', no ':' detected"),
    ("a:", 1, 1, None, "sheet", 2, 2, "Failed to decode code data 'a:', no '=' detected"),
    ("a:c", 1, 1, None, "sheet", 2, 2, "Failed to decode code data 'a:c', no '=' detected"),
    ("a:c=d", 1, 1, expected, "")
  ]
  for test in test_data:
    assert(base._decode_other_code(test[0], test[1], test[2])) == test[3]
    if not test[4] == "":
      mock_error.assert_called()
      assert mock_error.call_args[0][0] == test[4]
      assert mock_error.call_args[0][1] == test[5]
      assert mock_error.call_args[0][2] == test[6]
      assert mock_error.call_args[0][3] == test[7]

def test__state_split(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  base = BaseSheet("", globals, "sheet")
  test_data = [
    ('111', ['111']),
    ('111,    ', ['111']),
    ('"111", 222, 333', ['111', '222', '333']),
    ('"111", 222, 333, "4"', ['111', '222', '333', '4']),
    ('"111 \\" quote", 222, 333', ['111 " quote', '222', '333']),
    ('"111 \\" quote",,, 222, 333', ['111 " quote', '', '', '222', '333'])
  ]
  for test in test_data:
    assert(base._state_split(test[0])) == test[1]

  with pytest.raises(BaseSheet.FormatError):
    base._state_split('123, "456')

def test__to_int(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data)
  base = BaseSheet("", globals, "sheet")
  assert(base._to_int("4")) == 4
  assert(base._to_int("dd")) == None

def test_read_address_cell_by_name(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'address': [
    'line1, district, city, state, postal code, USA',
    'line1, line2, district, city, state, postal code, USA',
    'line1| line2| district| city| state| postal code| USA',
    ]
  }
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  mock_json = mocker.patch("json.load")
  mock_json.side_effect=[{}, {}, {}]
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="USA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1, expected_1, expected_1]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Addr_1', 'Addr_2', 'Addr_3']
  base = BaseSheet("", globals, "sheet")
  mock_ta = mocker.patch("usdm_excel.base_sheet.BaseSheet._to_address")
  mock_ta.side_effect=['success', 'success', 'success']
  assert base.read_address_cell_by_name(0, 'address') == 'success'
  assert base.read_address_cell_by_name(1, 'address') == 'success'
  assert base.read_address_cell_by_name(2, 'address') == 'success'
  assert mock_ta.call_count == 3
  expected = [
    mocker.call('Addr_1', lines=['line1'], district='district', city='city', state='state', postal_code='postal code', country=expected_1),
    mocker.call('Addr_2', lines=['line1', 'line2'], district='district', city='city', state='state', postal_code='postal code', country=expected_1),
    mocker.call('Addr_3', lines=['line1', 'line2'], district='district', city='city', state='state', postal_code='postal code', country=expected_1),
  ]
  mock_ta.assert_has_calls(expected)

def test_read_address_cell_by_name_errors(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'address': [
    'district, city, state, postal code, USA',
    '',
    ]
  }
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  mock_json = mocker.patch("json.load")
  mock_json.side_effect=[{}, {}, {}]
  expected_1 = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="USA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1, expected_1, expected_1]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Addr_1', 'Addr_2', 'Addr_3']
  base = BaseSheet("", globals, "sheet")
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  assert base.read_address_cell_by_name(0, 'address') == None
  assert base.read_address_cell_by_name(1, 'address') == None
  assert base.read_address_cell_by_name(1, 'address', True) == None
  assert mock_error.call_count == 2
  expected = [
    mocker.call('sheet', 1, 1, "Address 'district, city, state, postal code, USA' does not contain the required fields (lines, district, city, state, postal code and country code) using ',' separator characters, only 5 found", 40),
    mocker.call('sheet', 2, 1, "Address '' does not contain the required fields (lines, district, city, state, postal code and country code) using ',' separator characters, only 0 found", 40),
  ]
  mock_error.assert_has_calls(expected)

def test__to_address(mocker, globals):
  code = Code(id='Code1', code='code', codeSystem='codesys', codeSystemVersion='3', decode="USA")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {}
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  base = BaseSheet("", globals, "sheet")
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  result = base._to_address('Addr_1', ['line 1'], 'city', 'district', 'state', 'postal code', code)
  expected = {
    'city': 'city',
    'country': {'code': 'code',
               'codeSystem': 'codesys',
               'codeSystemVersion': '3',
               'decode': 'USA',
               'id': 'Code1',
               'instanceType': 'Code'},
    'district': 'district',
    'id': 'Addr_1',
    'instanceType': 'Address',
    'lines': ['line 1'],
    'postalCode': 'postal code',
    'state': 'state',
    'text': 'line 1, city, district, state, postal code, USA'
  }
  assert result.model_dump() == expected
  result = base._to_address('', ['line 1'], 'city', 'district', 'state', 'postal code', code)
  assert result == None
  assert mock_error.call_count == 1
  expected = [
    mocker.call('sheet', None, None, "Exception. Failed to create Address object. See log for additional details.", 40),
  ]
  mock_error.assert_has_calls(expected)

def test__read_geographic_scopes_cell(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'scope': [
    'Global',
    'Region: Europe',
    'Country: FRA',
    '',
    'Country=GBR'
    ]
  }
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  mock_json = mocker.patch("json.load")
  mock_json.side_effect=[{}, {}, {}]
  expected_1 = Code(id='Code1', code='code1', codeSystem='country', codeSystemVersion='2', decode="USA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1]
  expected_2 = Code(id='Code4', code='code4', codeSystem='region', codeSystemVersion='3', decode="Europe")
  mock_region_code = mocker.patch("usdm_excel.iso_3166.ISO3166.region_code")
  mock_region_code.side_effect=[expected_2]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Scope_1', 'Scope_2', 'Scope_3', 'Scope_4', 'Scope_5', 'Scope_6', 'Scope_7', 'Scope_8', 'Scope_9', 'Scope_10', 'Scope_11', 'Scope_12']
  base = BaseSheet("", globals, "sheet")
  assert base.read_geographic_scopes_cell(0, 0)[0].model_dump() == {
    'code': None, 
    'id': 'Scope_2', 
    'instanceType': 'GeographicScope', 
    'type': {'code': 'C68846', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Global', 'id': 'Scope_1', 'instanceType': 'Code'}
  }
  assert base.read_geographic_scopes_cell(1, 0)[0].model_dump() == {
    'code': {
      'id': 'Scope_4',
      'instanceType': 'AliasCode',
      'standardCode': {
          'code': 'code4',
          'codeSystem': 'region',
          'codeSystemVersion': '3',
          'decode': 'Europe',
          'id': 'Code4',
          'instanceType': 'Code',
      },
      'standardCodeAliases': []
    },
    'id': 'Scope_5', 
    'instanceType': 'GeographicScope', 
    'type': {'code': 'C41129', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Region', 'id': 'Scope_3', 'instanceType': 'Code'}
  }
  assert base.read_geographic_scopes_cell(2, 0)[0].model_dump() == {
    'code': {
      'id': 'Scope_7',
      'instanceType': 'AliasCode',
      'standardCode': {
          'code': 'code1',
          'codeSystem': 'country',
          'codeSystemVersion': '2',
          'decode': 'USA',
          'id': 'Code1',
          'instanceType': 'Code',
      },
      'standardCodeAliases': []
    },
    'id': 'Scope_8', 
    'instanceType': 'GeographicScope', 
    'type': {'code': 'C25464', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Country', 'id': 'Scope_6', 'instanceType': 'Code'}
  }
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  assert base.read_geographic_scopes_cell(3, 0)[0].model_dump() == {
    'code': None, 
    'id': 'Scope_10', 
    'instanceType': 'GeographicScope', 
    'type': {'code': 'C68846', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Global', 'id': 'Scope_9', 'instanceType': 'Code'}
  }
  assert base.read_geographic_scopes_cell(4, 0)[0].model_dump() == {
    'code': None, 
    'id': 'Scope_12', 
    'instanceType': 'GeographicScope', 
    'type': {'code': 'C68846', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Global', 'id': 'Scope_11', 'instanceType': 'Code'}
  }
  assert mock_error.call_count == 2
  expected = [
    mocker.call('sheet', 4, 1, "Empty cell detected where geographic scope values expected, assuming global scope.", 30),
    mocker.call('sheet', 5, 1, "Failed to decode geographic scope 'Country=GBR'. Formats are 'Global', 'Region: <value>' or 'Country: <value>'. Assuming global scope.", 30),
  ]
  mock_error.assert_has_calls(expected)
  
def test__read_geographic_scopes_cell_by_name(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {'region': [
    'Region: Europe',
    ]
  }
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  mock_json = mocker.patch("json.load")
  mock_json.side_effect=[{}]
  expected_1 = Code(id='Code1', code='code1', codeSystem='country', codeSystemVersion='2', decode="USA")
  mock_code = mocker.patch("usdm_excel.iso_3166.ISO3166.code")
  mock_code.side_effect=[expected_1]
  expected_2 = Code(id='Code4', code='code4', codeSystem='region', codeSystemVersion='3', decode="Europe")
  mock_region_code = mocker.patch("usdm_excel.iso_3166.ISO3166.region_code")
  mock_region_code.side_effect=[expected_2]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Scope_1', 'Scope_2', 'Scope_3', 'Scope_4', 'Scope_5', 'Scope_6', 'Scope_7', 'Scope_8', 'Scope_9']
  base = BaseSheet("", globals, "sheet")
  assert base.read_geographic_scopes_cell_by_name(0, 'region')[0].model_dump() == {
    'code': {
      'id': 'Scope_2',
      'instanceType': 'AliasCode',
      'standardCode': {
          'code': 'code4',
          'codeSystem': 'region',
          'codeSystemVersion': '3',
          'decode': 'Europe',
          'id': 'Code4',
          'instanceType': 'Code',
      },
      'standardCodeAliases': []
    },
    'id': 'Scope_3', 
    'instanceType': 'GeographicScope', 
    'type': {'code': 'C41129', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Region', 'id': 'Scope_1', 'instanceType': 'Code'}
  }
  assert base.read_geographic_scopes_cell_by_name(1, 'regionxxx')[0].model_dump() == {
    'code': None,
    'id': 'Scope_5', 
    'instanceType': 'GeographicScope', 
    'type': {'code': 'C68846', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Global', 'id': 'Scope_4', 'instanceType': 'Code'}
  }



from usdm_model.governance_date import GovernanceDate
from usdm_model.geographic_scope import GeographicScope
from usdm_model.study_title import StudyTitle
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.alias import Alias
from usdm_excel.cdisc_ct import CDISCCT
from usdm_excel.iso_3166 import ISO3166
from usdm_excel.option_manager import Options
from usdm_excel.globals import Globals

class StudySheet(BaseSheet):

  SHEET_NAME = 'study'
  
  NAME_TITLE = 'name'
  TITLE_TITLE = 'studyTitle'
  VERSION_TITLE = 'studyVersion'
  TYPE_TITLE = 'studyType'
  PHASE_TITLE = 'studyPhase'
  ACRONYM_TITLE = 'studyAcronym'
  RATIONALE_TITLE = 'studyRationale'
  TA_TITLE = 'businessTherapeuticAreas'
  BRIEF_TITLE_TITLE = 'briefTitle'
  OFFICAL_TITLE_TITLE = 'officialTitle'
  PUBLIC_TITLE_TITLE = 'publicTitle'
  SCIENTIFIC_TITLE_TITLE = 'scientificTitle'
  PROTOCOL_VERSION_TITLE = 'protocolVersion'
  PROTOCOL_STATUS_TITLE = 'protocolStatus'

  DATES_HEADER_ROW = 15
  DATES_DATA_ROW = 16
  
  PARAMS_NAME_COL = 0
  PARAMS_DATA_COL = 1

  STUDY_VERSION_DATE = 'study_version'
  PROTOCOL_VERSION_DATE = 'protocol_document'
  AMENDMENT_DATE = 'amendment'

  def __init__(self, file_path: str, globals: Globals):
    try:
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, header=None)
      self.date_categories = [self.STUDY_VERSION_DATE, self.PROTOCOL_VERSION_DATE, self.AMENDMENT_DATE]
      self.phase = None
      self.version = None
      self.type = None
      self.name = None
      self.brief_title = None
      self.official_title = None
      self.public_title = None
      self.scientific_title = None
      self.protocol_version = None
      self.protocol_status = None
      self.title = None
      self.titles = []
      self.acronym = None
      self.rationale = None
      self.study = None
      self.study_version = None
      self.protocol_document_version = None
      self.therapeutic_areas = []
      self.timelines = {}
      self.dates = {}
      for category in self.date_categories:
        self.dates[category] = []
      self._process_sheet()
    except Exception as e:
      self._sheet_exception(e)

  def _process_sheet(self):
    fields = ['category', 'name', 'description', 'label', 'type', 'date', 'scopes']    
    for rindex, row in self.sheet.iterrows():
      field_name = self.read_cell(rindex, self.PARAMS_NAME_COL)
      if field_name == self.NAME_TITLE:
        self.name = self.read_cell(rindex, self.PARAMS_DATA_COL)
      #elif field_name == self.TITLE_TITLE:
      #  if self.globals.option_manager.get(Options.USDM_VERSION) == '2':
      #    self.title = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif field_name == self.VERSION_TITLE:
        self.version = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif field_name == self.TYPE_TITLE:
        self.type = self.read_cdisc_klass_attribute_cell('Study', 'studyType', rindex, self.PARAMS_DATA_COL)
      elif field_name == self.PHASE_TITLE:
        phase = self.read_cdisc_klass_attribute_cell('Study', 'studyPhase', rindex, self.PARAMS_DATA_COL)
        self.phase = Alias(self.globals).code(phase, [])
      elif field_name == self.ACRONYM_TITLE:
        self.acronym = self._set_title(rindex, self.PARAMS_DATA_COL, "Study Acronym")
      elif field_name == self.RATIONALE_TITLE:
        self.rationale = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif field_name == self.TA_TITLE:
        self.therapeutic_areas = self.read_other_code_cell_mutiple(rindex, self.PARAMS_DATA_COL)
      elif field_name == self.BRIEF_TITLE_TITLE:
        self.brief_title = self._set_title(rindex, self.PARAMS_DATA_COL, "Brief Study Title")
      elif field_name == self.OFFICAL_TITLE_TITLE:
        self.official_title = self._set_title(rindex, self.PARAMS_DATA_COL, "Official Study Title")
      elif field_name == self.PUBLIC_TITLE_TITLE:
        self.public_title = self._set_title(rindex, self.PARAMS_DATA_COL, "Public Study Title")
      elif field_name == self.SCIENTIFIC_TITLE_TITLE:
        self.scientific_title = self._set_title(rindex, self.PARAMS_DATA_COL, "Scientific Study Title")
      elif field_name == self.PROTOCOL_VERSION_TITLE:
        self.protocol_version = self.read_cell(rindex, self.PARAMS_DATA_COL)
      elif field_name == self.PROTOCOL_STATUS_TITLE:
        self.protocol_status = self.read_cdisc_klass_attribute_cell('StudyProtocolVersion', 'protocolStatus', rindex, self.PARAMS_DATA_COL) 
      elif rindex >= self.DATES_DATA_ROW:
        record = {}
        for cindex in range(0, len(self.sheet.columns)):
          field = fields[cindex]
          if field == 'category':
            cell = self.read_cell(rindex, cindex)
            if cell.lower() in self.date_categories:
              category = cell.lower()
            else:
              categories = ', '.join(f'"{w}"' for w in self.date_categories)
              self._error(rindex, cindex, f"Date category not recognized, should be one of {categories}, defaults to '{self.date_categories[0]}'")
              category = self.date_categories[0]
          elif field == 'type':
            record[field] = self.read_cdisc_klass_attribute_cell('GovernanceDate', 'type', rindex, cindex)
          elif field == 'date':
            record[field] = self.read_date_cell(rindex, cindex)
          elif field == 'scopes':
            record[field] = self.read_geographic_scopes_cell(rindex, cindex)
          else:
            cell = self.read_cell(rindex, cindex)
            record[field] = cell
        try:
          date = GovernanceDate(
            id=self.globals.id_manager.build_id(GovernanceDate),
            name=record['name'],
            label=record['label'],
            description=record['description'],
            type=record['type'],
            dateValue=record['date'],
            geographicScopes=record['scopes']
          )
          self.dates[category].append(date)
          self.globals.cross_references.add(record['name'], date)
        except Exception as e:
          self._general_exception(f"Failed to create GovernanceDate object", e)

  def _set_title(self, rindex, cindex, title_type):
    # if self.globals.option_manager.get(Options.USDM_VERSION) == '2':
    #   return self.read_cell(rindex, cindex)
    # else:
    try:
      text = self.read_cell(rindex, cindex)
      if text:
        code = CDISCCT(self.globals).code_for_attribute('StudyVersion', 'titles', title_type)
        title = StudyTitle(id=self.globals.id_manager.build_id(StudyTitle), text=text, type=code)
        self.titles.append(title)
        self.globals.cross_references.add(title.id, title)
        return title
      else:
        return None
    except Exception as e:
      self._exception(rindex, cindex, "Failed to create StudyTitle object", e)
        


from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .alias_code import AliasCode
from .comment_annotation import CommentAnnotation

class BiomedicalConceptCategory(ApiBaseModelWithIdNameLabelAndDesc):
  childIds: List[str] = []
  memberIds: List[str] = []
  code: Union[AliasCode, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['BiomedicalConceptCategory']



from .utility import usdm_reference
from usdm_model.study_version import StudyVersion
from usdm_model.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.document.elements import Elements

class TemplateBase():

  def __init__(self, parent: BaseSheet, study_version: StudyVersion, document_version: StudyDefinitionDocumentVersion):
    self.parent = parent
    self._study_version = study_version
    self._study_design = self._study_version.studyDesigns[0]
    self._document_version = document_version
    self._elements = Elements(parent, study_version, document_version)
    self._methods = [func for func in dir(self.__class__) if callable(getattr(self.__class__, func)) and not func.startswith("_")]

  def valid_method(self, name):
    result = name in self._methods
    if not result:
      self.parent._general_warning(f"Could not resolve method name, {name} not in {self._methods}")
    return result
  
  def _reference(self, item, attribute):
    return usdm_reference(item, attribute)

  def _add_checking_for_tag(self, doc, tag, text):
    doc.asis(text)


from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging
from enum import Enum

class Options(Enum):
  EMPTY_NONE = 'empty_none'

class EmptyNoneOption(Enum):
  EMPTY = 'empty_string'
  NONE = 'none_value'

class OptionManager():

  def __init__(self, errors_and_logging: ErrorsAndLogging):
    self._errors_and_logging = errors_and_logging
    self._items = {}

  def clear(self):
    self._items = {}

  def set(self, name, value):
    name = self._to_string(name)
    value = self._to_string(value)
    self._items[name] = value

  def get(self, name):
    name = self._to_string(name)
    return self._items[name] if name in self._items else ''
  
  def _to_string(self, item):
    return item.value if isinstance(item, Enum) else str(item)
    




import pytest
import pandas as pd
from usdm_excel.study_design_conditions_sheet.study_design_conditions_sheet import StudyDesignConditionSheet
from usdm_model.api_base_model import ApiBaseModelWithId


xfail = pytest.mark.xfail

def test_create(mocker, globals):
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[
    ApiBaseModelWithId(id="X1"), ApiBaseModelWithId(id="X2"), ApiBaseModelWithId(id="X3"), ApiBaseModelWithId(id="X4"), ApiBaseModelWithId(id="X5"),
    ApiBaseModelWithId(id="X6"), ApiBaseModelWithId(id="X7"), ApiBaseModelWithId(id="X8"), ApiBaseModelWithId(id="X9"), ApiBaseModelWithId(id="X10")
  ]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Cond_1', 'Cond_2', 'Cond_3', 'Cond_4', 'Cond_5']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Id 1', 'Description One',   'Label One',   'Condition 1', 'Context 1', 'Applies 1'], 
    ['Id 2', 'Description Two',   'Label Two',   'Condition 2', 'Context 2', 'Applies 2'], 
    ['Id 3', 'Description Three', 'Label Three', 'Condition 3', 'Context 3', 'Applies 3'], 
    ['id 4', 'Description Four',  'Label Four',  'Condition 4', 'Context 4, context 5', 'Applies 4'], 
    ['Id 5', 'Description Five',  'Label Five',  'Condition 5', '',          'Applies 5']
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'text', 'context', 'appliesTo'])
  items = StudyDesignConditionSheet("", globals)
  assert len(items.items) == 5
  assert items.items[0].id == 'Cond_1'
  assert items.items[0].name == 'Id 1'
  assert items.items[0].description == 'Description One'
  assert items.items[0].label == 'Label One'
  assert items.items[0].contextIds == ['X1']
  assert items.items[0].appliesToIds == ['X2']
  assert items.items[3].contextIds == ['X7', 'X8']
  assert items.items[4].contextIds == []
  assert items.items[4].appliesToIds == ['X10']
  
def test_create_with_commas(mocker, globals):
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[
    ApiBaseModelWithId(id="X1"), ApiBaseModelWithId(id="X2"), ApiBaseModelWithId(id="X3"), ApiBaseModelWithId(id="X4"), ApiBaseModelWithId(id="X5"),
    ApiBaseModelWithId(id="X6"), ApiBaseModelWithId(id="X7"), ApiBaseModelWithId(id="X8"), ApiBaseModelWithId(id="X9"), ApiBaseModelWithId(id="X10")
  ]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Cond_1', 'Cond_2', 'Cond_3', 'Cond_4', 'Cond_5']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Id 1', 'Description One',   'Label One',   'Condition 1', 'Context 1', 'Applies 1'], 
    ['Id 2', 'Description Two',   'Label Two',   'Condition 2', 'Context 2', 'Applies 2'], 
    ['Id 3', 'Description Three', 'Label Three', 'Condition 3', 'Context 3', 'Applies 3'], 
    ['id 4', 'Description Four',  'Label Four',  'Condition 4', '"Context, 4", context 5', '"Applies, 4"'], 
    ['Id 5', 'Description Five',  'Label Five',  'Condition 5', '',          'Applies 5']
 ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'text', 'context', 'appliesTo'])
  items = StudyDesignConditionSheet("", globals)
  assert len(items.items) == 5
  assert items.items[0].id == 'Cond_1'
  assert items.items[0].name == 'Id 1'
  assert items.items[0].description == 'Description One'
  assert items.items[0].label == 'Label One'
  assert items.items[0].contextIds == ['X1']
  assert items.items[0].appliesToIds == ['X2']
  assert items.items[3].contextIds == ['X7', 'X8']
  assert items.items[3].appliesToIds == ['X9']
  assert items.items[4].contextIds == []
  assert items.items[4].appliesToIds == ['X10']
  
def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'text', 'context', 'appliesTo'])
  items = StudyDesignConditionSheet("", globals)
  assert len(items.items) == 0

def test_missing_reference(mocker, globals):
  
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  globals.cross_references.clear()
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[ApiBaseModelWithId(id="X1"), ApiBaseModelWithId(id="X2")]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Id 1', 'Description One',  'Condtition 1', 'Label One', 'Context 1', '']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'text', 'context', 'appliesTo'])
  items = StudyDesignConditionSheet("", globals)
  mock_error.assert_called()
  assert call_parameters == [
    ("studyDesignConditions", 1, 6, "No condition references found for '', at least one required", 40)
  ]

def test_read_cell_by_name_error(mocker, globals):
  
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  globals.cross_references.clear()
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[ApiBaseModelWithId(id="X1"), ApiBaseModelWithId(id="X2")]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['Id 1', 'Description One',   'Label One', 'Context 1', 'Applies 1']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'description', 'label', 'context', 'appliesTo'])
  items = StudyDesignConditionSheet("", globals)
  mock_error.assert_called()
  assert call_parameters == [
    ("studyDesignConditions", 1, -1, "Error attempting to read cell 'text'. Exception: Failed to detect column(s) 'text' in sheet", 40)
  ]
  



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .quantity import Quantity

class AdministrationDuration(ApiBaseModelWithId):
  quantity:	Union[Quantity, None] = None
  description: str
  durationWillVary: bool
  reasonDurationWillVary:	str
  instanceType: Literal['AdministrationDuration']



from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_v2_sheet.soa_column_rows import SoAColumnRows
from usdm_model.activity import Activity
from usdm_model.biomedical_concept_surrogate import BiomedicalConceptSurrogate
from usdm_model.procedure import Procedure
from usdm_model.schedule_timeline import ScheduleTimeline

class SoAActivity():
  
  def __init__(self, parent: BaseSheet, row_index: int, map: dict):
    self._parent_sheet = parent
    self._row_index = row_index
    self._map = map
    self.usdm_biomedical_concept_surrogates = []
    self.usdm_biomedical_concepts = []
    self._parent_name = parent.read_cell(row_index, SoAColumnRows.ACTIVITY_COL)
    self._child_name = parent.read_cell(row_index, SoAColumnRows.CHILD_ACTIVITY_COL)
    self._bcs, self._prs, self._tls = self._get_observation_cell(row_index, SoAColumnRows.BC_COL)
    self._parent_sheet._debug(row_index, SoAColumnRows.BC_COL, f"Activity {self._parent_name}, {self._child_name} read. BC: {self._bcs}, PR: {self._prs}, TL: {self._tls}")
    self.is_parent = False
    self.activity = self._process_parent() if self._parent_name else self._process_child()
    
  def _process_parent(self) -> Activity:
    self.is_parent = True
    if self._child_name:
      self._parent_sheet._warning(self._row_index, SoAColumnRows.CHILD_ACTIVITY_COL, f"Both parent '{self._parent_name}' and child activity '{self._child_name}' found, child has been ignored")
    if self._parent_name in self._map:
      self._parent_sheet._error(self._row_index, SoAColumnRows.ACTIVITY_COL, f"Parent activity '{self._parent_name}' has already been referenced in the SoA, parent has been ignored")
      return None
    else:
      return self._set_activity(self._parent_name, [], [], [], None)

  def _process_child(self) -> Activity:
    if self._child_name in self._map:
      self._parent_sheet._error(self._row_index, SoAColumnRows.CHILD_ACTIVITY_COL, f"Child activity '{self._child_name}' has already been referenced in the SoA, child has been ignored")
      return None
    else:
      full_bc_items, surrogate_bc_items = self._set_biomedical_concepts()
      timeline = self._set_timeline()
      procedures = self._set_procedures()
      return self._set_activity(self._child_name, full_bc_items, surrogate_bc_items, procedures, timeline)

  def _set_activity(self, name, full_bc_items, surrogate_bc_items, procedures, timeline):
    activity = self._parent_sheet.globals.cross_references.get(Activity, name)
    if activity is None:
      params = {'name': name, 'description': name, 'label': name, 'definedProcedures': procedures, 
                'biomedicalConceptIds': full_bc_items, 'bcCategoryIds': [], 
                'bcSurrogateIds': surrogate_bc_items, 'timelineId': timeline.id if timeline else None}
      activity = self._parent_sheet.create_object(Activity, params)
      if activity:
        self._parent_sheet.globals.cross_references.add(name, activity)     
        self._parent_sheet._warning(self._row_index, SoAColumnRows.BC_COL, f"No activity '{name}' found, so one has been created")
    else:
      activity.definedProcedures = procedures
      activity.biomedicalConceptIds = full_bc_items
      activity.bcSurrogateIds = surrogate_bc_items
      activity.timelineId = timeline.id if timeline else None
    self._map[activity.name] = activity
    return activity
  
  def _set_biomedical_concepts(self):
    full_bc_items = []
    surrogate_bc_items = []
    for bc in self._bcs:
      if self._parent_sheet.globals.cdisc_bc_library.exists(bc):
        full_bc = self._parent_sheet.globals.cdisc_bc_library.usdm(bc)
        full_bc_items.append(full_bc.id)
        self.usdm_biomedical_concepts.append(full_bc)
        self._parent_sheet.globals.cross_references.add(full_bc.id, full_bc)
      else:
        params = {'name': bc, 'description': bc, 'label': bc, 'reference': 'None set'}
        item = self._parent_sheet.create_object(BiomedicalConceptSurrogate, params)
        if item:
          surrogate_bc_items.append(item.id)
          self.usdm_biomedical_concept_surrogates.append(item)
          self._parent_sheet.globals.cross_references.add(item.id, item)
    return full_bc_items, surrogate_bc_items
  
  def _set_procedures(self):
    results = []
    for procedure in self._prs:
      ref = self._parent_sheet.globals.cross_references.get(Procedure, procedure)
      if ref:
        results.append(ref)
      else:
        self._parent_sheet._warning(self._row_index, SoAColumnRows.BC_COL, f"No procedure '{procedure}' found, missing cross reference")
    return results
  
  def _set_timeline(self):
    result = None
    if self._tls:
      result = self._parent_sheet.globals.cross_references.get(ScheduleTimeline, self._tls[0])
      if not result:
        self._parent_sheet._warning(self._row_index, SoAColumnRows.BC_COL, f"No timeline '{self._tls[0]}' found, missing cross reference")
    return result

  def _get_observation_cell(self, row_index: int, col_index: int) -> tuple[list, list, list]:
    bcs = []
    prs = []
    tls = []
    if not self._parent_sheet.cell_empty(row_index, col_index):
      value = self._parent_sheet.read_cell(row_index, col_index)
      outer_parts = value.split(',')
      for outer_part in outer_parts:
        parts = outer_part.split(':')
        if parts[0].strip().upper() == "BC":
          bcs.append(parts[1].strip())
        elif parts[0].strip().upper() == "PR":
          prs.append(parts[1].strip())
        elif parts[0].strip().upper() == "TL":
          tls.append(parts[1].strip())
        else:
          pass
    return bcs, prs, tls



from typing import Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code

class StudySite(ApiBaseModelWithIdNameLabelAndDesc):
  country: Code
  instanceType: Literal['StudySite']



import pandas as pd
from tests.test_factory import Factory
from usdm_excel.globals import Globals
from usdm_model.procedure import Procedure
from usdm_model.biomedical_concept import BiomedicalConcept
from usdm_model.schedule_timeline import ScheduleTimeline
from usdm_model.activity import Activity
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.study_soa_v2_sheet.soa_activity import SoAActivity

def test_read_bc(mocker, globals):
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[activities[0], bcs[0], None, None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 0, activity_map)
  assert item._bcs == ['BC1']
  assert item._prs == []
  assert item._tls == []
  assert item._parent_name == ''
  assert item._child_name == "Activity 1"
  usdm_activity = item.activity
  assert usdm_activity.name == "Activity 1"
  assert usdm_activity.label == 'Activity One'
  assert usdm_activity.description == None
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == ['BiomedicalConceptSurrogate_1']
  assert usdm_activity.definedProcedures == []
  assert usdm_activity.timelineId == None

def test_read_procedure(mocker, globals):
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[procedures[0], None, None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 1, activity_map)
  assert item._bcs == []
  assert item._prs == ['Procedure']
  assert item._tls == []
  assert item._parent_name == ''
  assert item._child_name == "Activity 2"
  usdm_activity = item.activity
  assert usdm_activity.name == "Activity 2"
  assert usdm_activity.label == "Activity 2"
  assert usdm_activity.description == "Activity 2"
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == []
  assert usdm_activity.definedProcedures == [procedures[0]]
  assert usdm_activity.timelineId == None

def test_read_procedure_error(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[None, activities[0], None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 1, activity_map)
  assert item._bcs == []
  assert item._prs == ['Procedure']
  assert item._tls == []
  assert item._parent_name == ''
  assert item._child_name == "Activity 2"
  usdm_activity = item.activity
  assert usdm_activity.name == "Activity 1"
  assert usdm_activity.label == 'Activity One'
  assert usdm_activity.description == None
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == []
  assert usdm_activity.definedProcedures == []
  assert usdm_activity.timelineId == None
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "sheet"
  assert mock_error.call_args[0][1] == 2
  assert mock_error.call_args[0][2] == 3
  assert mock_error.call_args[0][3] == "No procedure 'Procedure' found, missing cross reference"


def test_read_timeline(mocker, globals):
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[timelines[0], None, None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 2, activity_map)
  assert item._bcs == []
  assert item._prs == []
  assert item._tls == ['Timeline']
  assert item._parent_name == ''
  assert item._child_name == "Activity 3"
  usdm_activity = item.activity
  assert usdm_activity.name == "Activity 3"
  assert usdm_activity.label == "Activity 3"
  assert usdm_activity.description == "Activity 3"
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == []
  assert usdm_activity.definedProcedures == []
  assert usdm_activity.timelineId == 'ScheduleTimeline_1'

def test_read_timeline_error(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[None, activities[0], None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 2, activity_map)
  assert item._bcs == []
  assert item._prs == []
  assert item._tls == ['Timeline']
  assert item._parent_name == ''
  assert item._child_name == "Activity 3"
  usdm_activity = item.activity
  assert usdm_activity.name == "Activity 1"
  assert usdm_activity.label == 'Activity One'
  assert usdm_activity.description == None
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == []
  assert usdm_activity.definedProcedures == []
  assert usdm_activity.timelineId == None
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "sheet"
  assert mock_error.call_args[0][1] == 3
  assert mock_error.call_args[0][2] == 3
  assert mock_error.call_args[0][3] == "No timeline 'Timeline' found, missing cross reference"

def test_read_all(mocker, globals):
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[timelines[0], procedures[0], None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 3, activity_map)
  assert item._bcs == ['BC1']
  assert item._prs == ['Procedure']
  assert item._tls == ['Timeline']
  assert item._parent_name == ''
  assert item._child_name == "Activity 4"
  usdm_activity = item.activity
  assert usdm_activity.name == "Activity 4"
  assert usdm_activity.label == "Activity 4"
  assert usdm_activity.description == "Activity 4"
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == ['BiomedicalConceptSurrogate_1']
  assert usdm_activity.definedProcedures == [procedures[0]]
  assert usdm_activity.timelineId == 'ScheduleTimeline_1'

def test_read_parent(mocker, globals):
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[activities[2], None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 4, activity_map)
  assert item._bcs == ['BC1']
  assert item._prs == ['Procedure']
  assert item._tls == ['Timeline']
  assert item._parent_name == 'Parent 1'
  assert item._child_name == ''
  usdm_activity = item.activity
  assert usdm_activity.name == "Parent 1"
  assert usdm_activity.label == "Parent One"
  assert usdm_activity.description == None
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == []
  assert usdm_activity.definedProcedures == []
  assert usdm_activity.timelineId == None

def test_read_parent_ignore_child(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[None, None]
  base_sheet = _setup(mocker, globals)
  activity_map = {}
  item = SoAActivity(base_sheet, 5, activity_map)
  assert item._bcs == []
  assert item._prs == []
  assert item._tls == []
  assert item._parent_name == 'Parent 2'
  assert item._child_name == 'Activity 5'
  usdm_activity = item.activity
  assert usdm_activity.name == "Parent 2"
  assert usdm_activity.label == "Parent 2"
  assert usdm_activity.description == "Parent 2"
  assert usdm_activity.biomedicalConceptIds == []
  assert usdm_activity.bcCategoryIds == []
  assert usdm_activity.bcSurrogateIds == []
  assert usdm_activity.definedProcedures == []
  assert usdm_activity.timelineId == None
  assert mock_error.call_count == 2
  mock_error.assert_has_calls([mocker.call('sheet', 6, 2, "Both parent 'Parent 2' and child activity 'Activity 5' found, child has been ignored", 30), 
                               mocker.call("sheet", 6, 3, "No activity 'Parent 2' found, so one has been created", 30)])

def test_read_parent_repeated(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[None, None]
  base_sheet = _setup(mocker, globals)
  activity_map = {'Parent 1'}
  item = SoAActivity(base_sheet, 4, activity_map)
  assert item._bcs == ['BC1']
  assert item._prs == ['Procedure']
  assert item._tls == ['Timeline']
  assert item._parent_name == 'Parent 1'
  assert item._child_name == ''
  assert item.activity is None
  assert mock_error.call_count == 1
  mock_error.assert_has_calls([mocker.call('sheet', 5, 1, "Parent activity 'Parent 1' has already been referenced in the SoA, parent has been ignored", 40)])

def test_read_child_repeated(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  bcs, procedures, timelines, activities = _data(globals)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[None, None]
  base_sheet = _setup(mocker, globals)
  activity_map = {'Activity 1'}
  item = SoAActivity(base_sheet, 0, activity_map)
  assert item._bcs == ['BC1']
  assert item._prs == []
  assert item._tls == []
  assert item._parent_name == ''
  assert item._child_name == 'Activity 1'
  assert item.activity is None
  assert mock_error.call_count == 1
  mock_error.assert_has_calls([mocker.call('sheet', 1, 2, "Child activity 'Activity 1' has already been referenced in the SoA, child has been ignored", 40)])

def _setup(mocker, globals: Globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = {
    'col_1': ['', '', '', '', 'Parent 1', 'Parent 2'],
    'col_2': ['Activity 1', 'Activity 2', 'Activity 3', 'Activity 4', '', 'Activity 5'],
    'col_3': ['BC: BC1', 'PR: Procedure', 'TL: Timeline', 'BC: BC1, PR: Procedure, TL: Timeline', 'BC: BC1, PR: Procedure, TL: Timeline', '']
  }
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)
  return BaseSheet("", globals, "sheet")

def _data(globals: Globals):
  globals.id_manager.clear()
  factory = Factory(globals)
  item_list = [
    {'name': 'PR1', 'label': 'Procedure 1', 'description': 'Procedure One', 'procedureType': 'PR1', 'code': factory.cdisc_dummy()},
    {'name': 'PR2', 'label': 'Procedure 2', 'description': 'Procedure Two', 'procedureType': 'PR2', 'code': factory.cdisc_dummy()}
  ]
  procedures = factory.set(Procedure, item_list)

  item_list = [
    {'name': 'TL1', 'label': 'Timeline 1', 'description': 'Timeline One', 'mainTimeline': False, 'entryCondition': '', 'entryId': '', 'procedureType': 'PR1', 'code': factory.cdisc_dummy()},
    {'name': 'TL2', 'label': 'Timeline 2', 'description': 'Timeline Two', 'mainTimeline': False, 'entryCondition': '', 'entryId': '', 'procedureType': 'PR2', 'code': factory.cdisc_dummy()}
  ]
  timelines = factory.set(ScheduleTimeline, item_list)

  item_list = [
    {'name': 'BC1', 'label': 'Biomedicall Concept 1', 'reference': 'BC REF 1', 'code': factory.alias_code(factory.cdisc_dummy())},
    {'name': 'BC2', 'label': 'Biomedicall Concept 2', 'reference': 'BC REF 2', 'code': factory.alias_code(factory.cdisc_dummy())}
  ]
  bcs = factory.set(BiomedicalConcept, item_list)

  item_list = [
    {'name': 'Activity 1', 'label': 'Activity One'},
    {'name': 'Activity 2', 'label': 'Activity Two'},
    {'name': 'Parent 1', 'label': 'Parent One'}
  ]
  activities = factory.set(Activity, item_list)

  return bcs, procedures, timelines, activities
  


from usdm_excel.base_sheet import BaseSheet
from usdm_model.assigned_person import AssignedPerson
from usdm_model.organization import Organization
from usdm_excel.globals import Globals

class AssignedPersonSheet(BaseSheet):

  SHEET_NAME = 'people'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        for index, row in self.sheet.iterrows():
          params = {
            'name': self.read_cell_by_name(index, 'name'), 
            'description': self.read_cell_by_name(index, 'description', default=''), 
            'label': self.read_cell_by_name(index, 'label', default=''), 
            'jobTitle': self.read_cell_by_name(index, 'jobTitle'), 
            'organizationId': self._get_org(index)}
          item = self.create_object(AssignedPerson, params)
          if item:
            self.items.append(item)
            self.globals.cross_references.add(item.name, item)     
    except Exception as e:
      self._sheet_exception(e)

  def _get_org(self, index: int) -> str:
    org_ref = self.read_cell_by_name(index, 'organization', default='')
    org = self.globals.cross_references.get(Organization, org_ref) if org_ref else None
    return org.id if org else None



import pytest
import pandas as pd
from usdm_excel.study_definition_document.document_template_sheet import DocumentTemplateSheet
from usdm_excel.option_manager import Options, EmptyNoneOption
from usdm_model.api_base_model import ApiBaseModelWithIdAndName

xfail = pytest.mark.xfail

def test_create(mocker, globals):
  globals.option_manager.set(Options.EMPTY_NONE, EmptyNoneOption.EMPTY)
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[
    ApiBaseModelWithIdAndName(id="X1",  name='ITEM1'), 
    ApiBaseModelWithIdAndName(id="X2",  name='ITEM2'), 
    ApiBaseModelWithIdAndName(id="X3",  name='ITEM3'), 
    ApiBaseModelWithIdAndName(id="X4",  name='ITEM4'), 
    ApiBaseModelWithIdAndName(id="X5",  name='ITEM5'),
    ApiBaseModelWithIdAndName(id="X6",  name='ITEM6'), 
    ApiBaseModelWithIdAndName(id="X7",  name='ITEM7'), 
    ApiBaseModelWithIdAndName(id="X8",  name='ITEM8'), 
    ApiBaseModelWithIdAndName(id="X9",  name='ITEM9'), 
    ApiBaseModelWithIdAndName(id="X10", name='ITEM10')
  ]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['CONTENT1', '1',     True, 'Section 1',     True, 'ITEM1'], 
    ['CONTENT2', '1.1',   True, 'Section 1.1',   True, 'ITEM2'], 
    ['CONTENT3', '1.2',   True, 'Section 1.2',   True, 'ITEM3'], 
    ['CONTENT4', '1.2.1', True, 'Section 1.2.1', True, 'ITEM4'], 
    ['CONTENT5', '2',     True, 'Section 2',     True, 'ITEM5'], 
    ['CONTENT6', '2.1',   True, 'Section 2.1',   True, 'ITEM6'], 
    ['CONTENT7', '3',     True, 'Section 3',     True, 'ITEM7']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'sectionNumber', 'displaySectionNumber', 'sectionTitle', 'displaySectionTitle', 'content'])
  content = DocumentTemplateSheet("", "Template", "", globals)
  assert content.name == "Template"
  assert len(content.items) == 7
  assert content.items[0].name == 'CONTENT1'
  assert content.items[0].previousId == ''
  assert content.items[0].nextId == 'Content_2'  
  assert content.items[0].childIds == ['Content_2', 'Content_3']
  assert content.items[1].id == 'Content_2'
  assert content.items[1].name == 'CONTENT2'
  assert content.items[1].sectionNumber == '1.1'
  assert content.items[1].sectionTitle == 'Section 1.1'
  assert content.items[1].previousId == 'Content_1'
  assert content.items[1].nextId == 'Content_3'  
  assert content.items[1].childIds == []
  assert content.items[2].id == 'Content_3'
  assert content.items[2].name == 'CONTENT3'
  assert content.items[2].previousId == 'Content_2'
  assert content.items[2].nextId == 'Content_4'  
  assert content.items[2].childIds == ['Content_4']
  assert content.items[3].id == 'Content_4'
  assert content.items[3].name == 'CONTENT4'
  assert content.items[3].previousId == 'Content_3'
  assert content.items[3].nextId == 'Content_5'  
  assert content.items[3].childIds == []
  assert content.items[4].id == 'Content_5'
  assert content.items[4].name == 'CONTENT5'
  assert content.items[4].childIds == ['Content_6']
  assert content.items[5].id == 'Content_6'
  assert content.items[5].name == 'CONTENT6'
  assert content.items[5].sectionNumber == '2.1'
  assert content.items[5].sectionTitle == 'Section 2.1'
  assert content.items[5].childIds == []
  assert content.items[6].name == 'CONTENT7'
  assert content.items[6].id == 'Content_7'
  assert content.items[6].sectionNumber == '3'
  assert content.items[6].childIds == []

# def test_create_training_dot(mocker, globals):
#   mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
#   mock_present.side_effect=[True]
#   mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
#   mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
#   mocked_open = mocker.mock_open(read_data="File")
#   mocker.patch("builtins.open", mocked_open)
#   data = [
#     ['1',     '',         'Section 1',     'Text 1'], 
#     ['1.1',   'SET NAME', 'Section 1.1',   'Text 1.1'], 
#     ['1.2',   '',         'Section 1.2',   'Text 1.2'], 
#     ['1.2.1', '',         'Section 1.2.1', 'Text 1.2.1.'], 
#     ['2',     '',         'Section 2',     'Text 2'], 
#     ['2.1',   '',         'Section 2.1',   'Text 2.1.'], 
#     ['3',     '',         'Section 3',     'Text 3'], 
#   ]
#   mock_read = mocker.patch("pandas.read_excel")
#   mock_read.return_value = pd.DataFrame(data, columns=['name', 'sectionNumber', 'displaySectionNumber', 'sectionTitle', 'displaySectionTitle', 'content'])
#   content = DocumentTemplateSheet("", "", globals)
#   assert len(content.items) == 8
#   assert content.items[0].name == 'ROOT'
#   assert content.items[1].id == 'Content_2'
#   assert content.items[1].name == 'SECTION 1'
#   assert content.items[1].sectionNumber == '1'
#   assert content.items[1].sectionTitle == 'Section 1'
#   assert content.items[1].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 1</div>'
#   assert content.items[1].childIds == ['Content_3', 'Content_4']
#   assert content.items[2].name == 'SET NAME'
#   assert content.items[3].name == 'SECTION 1.2'
#   assert content.items[4].name == 'SECTION 1.2.1'
#   assert content.items[5].name == 'SECTION 2'
#   assert content.items[6].name == 'SECTION 2.1'
#   assert content.items[7].id == 'Content_8'
#   assert content.items[7].name == 'SECTION 3'

def test_create_4_levels(mocker, globals):
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[
    ApiBaseModelWithIdAndName(id="X1",  name='ITEM1'), 
    ApiBaseModelWithIdAndName(id="X2",  name='ITEM2'), 
    ApiBaseModelWithIdAndName(id="X3",  name='ITEM3'), 
    ApiBaseModelWithIdAndName(id="X4",  name='ITEM4'), 
    ApiBaseModelWithIdAndName(id="X5",  name='ITEM5'),
    ApiBaseModelWithIdAndName(id="X6",  name='ITEM6'), 
    ApiBaseModelWithIdAndName(id="X7",  name='ITEM7'), 
    ApiBaseModelWithIdAndName(id="X8",  name='ITEM8'), 
    ApiBaseModelWithIdAndName(id="X9",  name='ITEM9'), 
    ApiBaseModelWithIdAndName(id="X10", name='ITEM10')
  ]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['CONTENT1', '1',       True, 'Section 1',       True, 'ITEM1'], 
    ['CONTENT2', '1.1',     True, 'Section 1.1',     True, 'ITEM2'], 
    ['CONTENT3', '1.2',     True, 'Section 1.2',     True, 'ITEM3'], 
    ['CONTENT4', '1.2.1',   True, 'Section 1.2.1',   True, 'ITEM4'], 
    ['CONTENT5', '1.2.1.1', True, 'Section 1.2.1.1', True, 'ITEM5'], 
    ['CONTENT6', '2',       True, 'Section 2',       True, 'ITEM6'], 
    ['CONTENT7', '2.1',     True, 'Section 2.1',     True, 'ITEM7'], 
    ['CONTENT8', '3',       True, 'Section 3',       True, 'ITEM8']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'sectionNumber', 'displaySectionNumber', 'sectionTitle', 'displaySectionTitle', 'content'])
  content = DocumentTemplateSheet("", "Template", "", globals)
  assert len(content.items) == 8
  assert content.items[3].id == 'Content_4'
  assert content.items[3].name == 'CONTENT4'
  assert content.items[3].sectionNumber == '1.2.1'
  assert content.items[3].sectionTitle == 'Section 1.2.1'
  assert content.items[3].childIds == ['Content_5']
  assert content.items[4].id == 'Content_5'
  assert content.items[4].name == 'CONTENT5'
  assert content.items[4].sectionNumber == '1.2.1.1'
  assert content.items[4].displaySectionNumber == True
  assert content.items[4].sectionTitle == 'Section 1.2.1.1'
  assert content.items[4].displaySectionTitle == True
  assert content.items[4].childIds == []

# def test_create_standard_section(mocker, globals):
#   mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
#   mock_present.side_effect=[True]
#   mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
#   mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
#   mocked_open = mocker.mock_open(read_data="File")
#   mocker.patch("builtins.open", mocked_open)
#   data = [
#     ['1',       '',         'Section 1',       'Text 1'], 
#     ['1.1',     'SET NAME', 'Section 1.1',     'Text 1.1'], 
#     ['1.2',     '',         'Section 1.2',     '<usdm:section name="m11-title">'], 
#     ['1.2.1',   '',         'Section 1.2.1',   '<div><usdm:section name="m11-title"></div>'], 
#     ['2',       '',         'Section 2',       'Text 2'], 
#   ]
#   mock_read = mocker.patch("pandas.read_excel")
#   mock_read.return_value = pd.DataFrame(data, columns=['sectionNumber', 'name', 'sectionTitle', 'text'])
#   content = DocumentTemplateSheet("", globals)
#   assert len(content.items) == 6
#   assert content.items[0].text == ''
#   assert content.items[1].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 1</div>'
#   assert content.items[2].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 1.1</div>'
#   assert content.items[3].text == '<div xmlns="http://www.w3.org/1999/xhtml"><usdm:section name="m11-title"></div>'
#   assert content.items[4].text == '<div xmlns="http://www.w3.org/1999/xhtml"><usdm:section name="m11-title"></div>'
#   assert content.items[5].text == '<div xmlns="http://www.w3.org/1999/xhtml">Text 2</div>'

def test_create_invalid_levels(mocker, globals):
  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[
    ApiBaseModelWithIdAndName(id="X1",  name='ITEM1'), 
    ApiBaseModelWithIdAndName(id="X2",  name='ITEM2'), 
    ApiBaseModelWithIdAndName(id="X3",  name='ITEM3'), 
    ApiBaseModelWithIdAndName(id="X4",  name='ITEM4'), 
    ApiBaseModelWithIdAndName(id="X5",  name='ITEM5'),
    ApiBaseModelWithIdAndName(id="X6",  name='ITEM6'), 
    ApiBaseModelWithIdAndName(id="X7",  name='ITEM7'), 
    ApiBaseModelWithIdAndName(id="X8",  name='ITEM8'), 
    ApiBaseModelWithIdAndName(id="X9",  name='ITEM9'), 
    ApiBaseModelWithIdAndName(id="X10", name='ITEM10')
  ]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Content_1', 'Content_2', 'Content_3', 'Content_4', 'Content_5', 'Content_6', 'Content_7', 'Content_8']
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    ['CONTENT1', '1',         True, 'Section 1',       True, 'ITEM1'], 
    ['CONTENT2', '1.1',       True, 'Section 1.1',     True, 'ITEM2'], 
    ['CONTENT3', '1.2',       True, 'Section 1.2',     True, 'ITEM3'], 
    ['CONTENT4', '1.2.1',     True, 'Section 1.2.1',   True, 'ITEM4'], 
    ['CONTENT5', '1.2.1.1.4', True, 'Section 1.2.1.1', True, 'ITEM5'], 
    ['CONTENT6', '2',         True, 'Section 2',       True, 'ITEM6'], 
    ['CONTENT7', '2.1',       True, 'Section 2.1',     True, 'ITEM7'], 
    ['CONTENT8', '3',         True, 'Section 3',       True, 'ITEM8']
  ]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'sectionNumber', 'displaySectionNumber', 'sectionTitle', 'displaySectionTitle', 'content'])
  DocumentTemplateSheet("", "Template", "XXX", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "XXX"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception. Error [] while reading sheet 'XXX'. See log for additional details."

def test_create_empty(mocker, globals):
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'sectionNumber', 'displaySectionNumber', 'sectionTitle', 'displaySectionTitle', 'content'])
  content = DocumentTemplateSheet("", "Template", "", globals)
  assert len(content.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  globals.cross_references.clear()
  globals.template_manager.add('sponsor', 'studyDesignContent')
  call_parameters = []
  
  def my_add(sheet, row, column, message, level=10):
    call_parameters.append((sheet, row, column, message, level))
    return None

  mock_cross_ref = mocker.patch("usdm_excel.cross_ref.CrossRef.get")
  mock_cross_ref.side_effect=[ApiBaseModelWithIdAndName(id="X1",  name='ITEM1')]
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add", side_effect=my_add)
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['CONTENT1', '1', True, True, 'ITEM1']]
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=['name', 'sectionNumber', 'displaySectionNumber', 'displaySectionTitle', 'content'])
  content = DocumentTemplateSheet("", "Template", "XXX", globals)
  mock_error.assert_called()
  assert call_parameters == [
    ('XXX', 1, -1, "Error attempting to read cell 'sectionTitle'. Exception: Failed to detect column(s) 'sectionTitle' in sheet", 40)
  ]
  


from usdm_excel.cdisc_ct_library import CDISCCTLibrary

def test_load(mocker, globals):
    item = CDISCCTLibrary(globals.errors_and_logging)
    assert(item.submission('E-MAIL'))['conceptId'] == 'C25170' 
    assert(item.preferred_term('E-mail'))['conceptId'] == 'C25170' 


import traceback
import pandas as pd
from usdm_excel.base_sheet import BaseSheet
from usdm_model.study_element import StudyElement
from usdm_model.transition_rule import TransitionRule
from usdm_excel.globals import Globals

class StudyDesignElementSheet(BaseSheet):

  SHEET_NAME = 'studyDesignElements'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      for index, row in self.sheet.iterrows():
        start_rule = None
        end_rule = None
        xref = self.read_cell_by_name(index, 'xref', default="", must_be_present=False)
        name = self.read_cell_by_name(index, ['studyElementName', 'name'])
        description = self.read_cell_by_name(index, ['studyElementDescription', 'description'])
        label = self.read_cell_by_name(index, 'label', default="", must_be_present=False)
        start_rule_text = self.read_cell_by_name(index, 'transitionStartRule')
        end_rule_text = self.read_cell_by_name(index, 'transitionEndRule')
        if start_rule_text:
          start_rule = TransitionRule(id=self.globals.id_manager.build_id(TransitionRule), name=f"ELEMENT_START_RULE_{index + 1}", text=start_rule_text)
        if end_rule_text:
          end_rule = TransitionRule(id=self.globals.id_manager.build_id(TransitionRule), name=f"ELEMENT_END_RULE_{index + 1}", text=end_rule_text)
        try:
          item = StudyElement(
            id=self.globals.id_manager.build_id(StudyElement), 
            name=name,
            description=description,
            label=label,
            transitionStartRule=start_rule,
            transitionEndRule=end_rule
          )
        except Exception as e:
          self._general_exception(f"Failed to create StudyElement object", e)
        else:
          self.items.append(item)
          cross_ref = xref if xref else name
          self.globals.cross_references.add(cross_ref, item)     
    except Exception as e:
      self._sheet_exception(e)




import traceback
import pandas as pd
from usdm_excel.base_sheet import BaseSheet
from usdm_model.encounter import Encounter
from usdm_model.transition_rule import TransitionRule
from usdm_model.timing import Timing
from usdm_excel.globals import Globals

class StudyDesignEncounterSheet(BaseSheet):

  SHEET_NAME = 'studyDesignEncounters'

  def __init__(self, file_path: str, globals: Globals):
    try:
      self.items = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      for index, row in self.sheet.iterrows():
        start_rule = None
        end_rule = None
        xref = self.read_cell_by_name(index, 'xref', must_be_present=False)
        name = self.read_cell_by_name(index, ['encounterName', 'name'])
        description = self.read_cell_by_name(index, ['encounterDescription', 'description'])
        label = self.read_cell_by_name(index, 'label', default='', must_be_present=False)
        type = self.read_cdisc_klass_attribute_cell_by_name('Encounter', 'encounterType', index, ['encounterType', 'type'])
        settings = self.read_cdisc_klass_attribute_cell_multiple_by_name('Encounter', 'encounterEnvironmentalSetting', index, ['encounterEnvironmentalSetting', 'environmentalSetting'])
        modes = self.read_cdisc_klass_attribute_cell_multiple_by_name('Encounter', 'encounterContactModes', index, ['encounterContactModes', 'contactModes'])
        start_rule_text = self.read_cell_by_name(index, 'transitionStartRule')
        end_rule_text = self.read_cell_by_name(index, 'transitionEndRule')
        timing_xref = self.read_cell_by_name(index, 'window', must_be_present=False)
        if start_rule_text:
          start_rule = TransitionRule(id=self.globals.id_manager.build_id(TransitionRule), name=f"ENCOUNTER_START_RULE_{index + 1}", text=start_rule_text)
        if end_rule_text:
          end_rule = TransitionRule(id=self.globals.id_manager.build_id(TransitionRule), name=f"ENCOUNTER_START_RULE_{index + 1}", text=end_rule_text)
        if timing_xref:
          timing = self.globals.cross_references.get(Timing, timing_xref)
          timing_id = timing.id if timing else None
        else:
          timing_id = None
        try:
          item = Encounter(
            id=self.globals.id_manager.build_id(Encounter), 
            name=name,
            description=description,
            label=label,
            type=type, 
            environmentalSettings=settings,
            contactModes=modes,
            transitionStartRule=start_rule,
            transitionEndRule=end_rule,
            scheduledAtId=timing_id
          )
        except Exception as e:
          self._general_exception(f"Failed to create Encounter object", e)
        else:
          self.items.append(item)
          cross_ref = xref if xref else name
          self.globals.cross_references.add(cross_ref, item)     
      self.double_link(self.items, 'previousId', 'nextId')   
    except Exception as e:
      self._sheet_exception(e)




from usdm_excel.globals import Globals
from usdm_model.code import Code
from usdm_excel.code_base import CodeBase

class NCIt(CodeBase):

  def __init__(self, globals: Globals):
    super().__init__(globals)

  def code(self, code, decode):
    return Code(id=self._id_manager.build_id(Code), code=code, codeSystem='NCI Thesaurus', codeSystemVersion='24.01e', decode=decode)




from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .document_content_reference import DocumentContentReference

class StudyChange(ApiBaseModelWithIdNameLabelAndDesc):
  summary: str
  rationale: str
  changedSections: List[DocumentContentReference]
  instanceType: Literal['StudyChange']



import re
from yattag import Doc
from .template_base import TemplateBase
from usdm_excel.base_sheet import BaseSheet
from usdm_model.study import Study

class TemplateM11(TemplateBase):

  # def __init__(self, parent: BaseSheet, study: Study):
  #   super().__init__(parent, study)

  def title_page(self, attributes):
    doc = Doc()
    with doc.tag('table'):
      self._title_page_entry(doc, 'Sponsor Confidentiality Statement:', '')
      self._title_page_entry(doc, 'Full Title:', f'{self._elements.study_full_title()}')
      self._title_page_entry(doc, 'Trial Acronym:', f'{self._elements.study_acronym()}')
      self._title_page_entry(doc, 'Protocol Identifier:', f'{self._elements.study_identifier()}')
      self._title_page_entry(doc, 'Original Protocol:', '')
      self._title_page_entry(doc, 'Version Number:', f'{self._elements.study_version_identifier()}')
      self._title_page_entry(doc, 'Version Date:', f'{self._elements.study_approval_date()}')
      self._title_page_entry(doc, 'Amendment Identifier:', f'{self._elements.amendment()}')
      self._title_page_entry(doc, 'Amendment Scope:', f'{self._elements.amendment_scopes()}')
      self._title_page_entry(doc, 'Compound Codes(s):', '')
      self._title_page_entry(doc, 'Compound Name(s):', '')
      self._title_page_entry(doc, 'Trial Phase:', f'{self._elements.study_phase()}')
      self._title_page_entry(doc, 'Short Title:', f'{self._elements.study_short_title()}')
      self._title_page_entry(doc, 'Sponsor Name and Address:', f'{self._elements.organization_name_and_address()}')
      self._title_page_entry(doc, 'Regulatory Agency Identifier Number(s):', f'{self._elements.study_regulatory_identifiers()}')
      self._title_page_entry(doc, 'Spondor Approval Date:', f'{self._elements.document_approval_date()}')

      # Enter Nonproprietary Name(s)
      # Enter Proprietary Name(s)
      # Globally/Locally/Cohort
      # Primary Reason for Amendment
      # Region Identifier
      # Secondary Reason for Amendment

    result = doc.getvalue()
    #print(f"DOC: {result}")
    return result

  def inclusion(self, attributes):  
    return self._criteria("C25532")
  
  def exclusion(self, attributes):  
    return self._criteria("C25370")

  def objective_endpoints(self, attributes):
    #print(f"M11 TP:")
    doc = Doc()
    with doc.tag('table'):
      for item in self._objective_endpoints_list():
        self._objective_endpoints_entry(doc, item['objective'], item['endpoints'])
    return doc.getvalue()

  def _criteria(self, type):
    #print(f"M11 TP:")
    heading = { 
      'C25532': "Patients may be included in the study only if they meet <strong>all</strong> the following criteria:",
      'C25370': "Patients may be excluded in the study for <strong>any</strong> of the following reasons:",
    }
    doc = Doc()
    with doc.tag('p'):
      doc.asis(heading[type])  
    with doc.tag('table'):
      for criterion in self._criteria_list(type):
        self._critieria_entry(doc, criterion['identifier'], criterion['text'])
    return doc.getvalue()

  def _critieria_entry(self, doc, number, entry):
    with doc.tag('tr'):
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(number)  
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(entry)

  def _objective_endpoints_entry(self, doc, objective, endpoints):
    with doc.tag('tr'):
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(objective)  
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        for endpoint in endpoints:
          with doc.tag('p'):
            doc.asis(endpoint)

  def _title_page_entry(self, doc, title, entry):
    with doc.tag('tr'):
      with doc.tag('th', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(title)  
      with doc.tag('td', style="vertical-align: top; text-align: left"):
        with doc.tag('p'):
          doc.asis(entry)

  def _criteria_list(self, type):
    results = []
    criteria = {x.id:x for x in self._study_version.criteria}
    items = [criteria[c] for c in self._study_design.population.criterionIds if criteria[c].category.code == type ]
    items.sort(key=lambda d: d.identifier)
    for item in items:
      result = {'identifier': item.identifier, 'text': self._reference(item, 'text')}
      results.append(result)
    return results

  def _objective_endpoints_list(self):
    results = []
    for item in self._study_design.objectives:
      result = {'objective': self._reference(item, 'text'), 'endpoints': []}
      for endpoint in item.endpoints:
        result['endpoints'].append(self._reference(endpoint, 'text'))
      results.append(result)
    return results
  


from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging

class BaseManager():

  def __init__(self, errors_and_logging: ErrorsAndLogging):
    self._errors_and_logging = errors_and_logging
    self._items = {}

  def __iter__(self):
    return iter(self._items)

  def clear(self) -> None:
    self._items = {}

  def add(self, name: str, value: str) -> None:
    self._items[name.upper()] = value

  def get(self, name: str) -> str:
    u_name = name.upper()
    return self._items[u_name] if u_name in self._items else ''
  
  def includes(self, name) -> bool:
    u_name = name.upper()
    return u_name in self._items
  
  # def all(self) -> list:
  #   return list(self._items.values())
  
  def items(self):
    return self._items.items()
  
  def keys(self):
    return self._items.keys()

  def values(self):
    return self._items.values()


import pytest
import json
import pandas as pd
from usdm_excel.study_design_intervention_sheet.study_design_intervention_sheet import StudyDesignInterventionSheet
from usdm_model.code import Code


xfail = pytest.mark.xfail

SAVE = True
COLUMNS = [ 'name', 'description', 'label', "codes", "role", "type", 
  "pharmacologicalClass", "productDesignation", "minimumResponseDuration", 'administrationName', 'administrationDescription', 
  'administrationLabel', "administrationRoute", "administrationDose", "administrationFrequency", 
  'administrationDurationDescription', 'administrationDurationWillVary', 'administrationDurationWillVaryReason', 'administrationDurationQuantity' ]

def read_json(filename):
  with open(filename, 'r') as f:
    data = json.load(f)
  return data

def dump_json(data):
  print(f"\n{json.dumps(data, indent=2)}")

def test_create(mocker, globals):
  expected = read_json(f"tests/integration_test_files/intervention/create.json")
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=[f"Id_{x}" for x in range(100)]

  mock_code = mocker.patch("usdm_excel.cdisc_ct.CDISCCT.code_for_attribute")
  mock_code.side_effect=[Code(id=f'Code_{x}', code='C12345', codeSystem='CDISC', codeSystemVersion='1', decode=f"INDEX{x}") for x in range(100)]
  
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [
    # name     description    label          codes                          role                         type      pharmacologicalClass  productDesignation      minimumResponseDuration,  administrationName [10] administrationDescription administrationLabel administrationRoute administrationDose administrationFrequency administrationDurationDescription administrationDurationWillVary administrationDurationWillVaryReason administrationDurationQuantity
    [ 'Int 1', 'Int Desc 1',  'Int Label 1', 'SPONSOR: A=B',                'Experimental Intervention', 'C12345', 'FDA: A=B',           'IMP',                  '1 Day',                  'Admin 1',              'Admin Desc 1',           'Admin Label 1',    'C34567',          '12 mg',            'C65432',               'Dur desc 1',                     'False',                       '',                                  '14 %'                         ], 
    [ '',      '',            '',            '',                            '',                          '',       '',                   '',                     '',                       'Admin 2',              'Admin Desc 2',           'Admin Label 1',    'C34567',          '1 mg',             'C65432',               'Dur desc 1',                     'False',                       '',                                  '10 m'                         ], 
    [ 'Int 2', 'Int Desc 2',  'Int Label 2', 'SPONSOR: C=D',                'Placebo',                   'C12345', 'FDA: A=B',           'NIMP (AxMP)',          '3 Weeks',                'Admin 3',              'Admin Desc 3',           'Admin Label 1',    'C34567',          '100 mg',           'C65432',               'Dur desc 1',                     'False',                       '',                                  '12 C'                         ], 
    [ 'Int 3', 'Int Desc 3',  'Int Label 3', 'SPONSOR: E=F, SPONSOR: G=H',  'Rescue Medicine',           'C12345', 'FDA: A=B',           'IMP',                  '4 Years',                'Admin 4',              'Admin Desc 4',           'Admin Label 1',    'C34567',          '500 mg',           'C65432',               'Dur desc 1',                     'False',                       '',                                  '12 F'                         ], 
    [ '',      '',            '',            '',                            '',                          '',       '',                   '',                     '',                       'Admin 5',              'Admin Desc 5',           'Admin Label 1',    'C34567',          '1 mg',             'C65432',               'Dur desc 1',                     'False',                       '',                                  '15 in'                        ], 
  ]

  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=COLUMNS)
  interventions = StudyDesignInterventionSheet("", globals)
  #dump_json({'items': [json.loads(x.to_json()) for x in interventions.items]})
  for index, item in enumerate(interventions.items):
    result = json.loads(item.to_json())
    assert result == expected["items"][index]
  
def test_create_empty(mocker, globals):
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = []
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame(data, columns=COLUMNS)
  interventions = StudyDesignInterventionSheet("", globals)
  assert len(interventions.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  data = [['', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '']]
  mock_read = mocker.patch("pandas.read_excel")
  columns = COLUMNS
  columns = columns[0:-1]
  mock_read.return_value = pd.DataFrame(data, columns=columns)
  interventions = StudyDesignInterventionSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "studyDesignInterventions"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception. Error [Failed to detect column(s) 'administrationDurationQuantity' in sheet] while reading sheet 'studyDesignInterventions'. See log for additional details."
  



from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging

class IdManager():

  def __init__(self, errors_and_logging: ErrorsAndLogging):
    self._errors_and_logging = errors_and_logging
    self._id_index = {}
    self.clear()

  def clear(self):
    self._id_index = {
      'Address': 0,
      'Code': 0,
      'AliasCode': 0,
      'Organization': 0,
      'StudyIdentifier': 0,
      'StudyEpoch': 0,
      'StudyArm': 0,
      'StudyCell': 0,
      'Entry': 0, 
      'ScheduleTimelineExit': 0, 
      'ScheduledActivityInstance': 0,
      'ScheduledDecisionInstance': 0,
      'ScheduleTimeline': 0, 
      'Timing': 0, 
      'StudyDesign': 0, 
      'Study': 0, 
      'StudyVersion': 0, 
      'Activity': 0, 
      'Encounter': 0, 
      'BiomedicalConceptSurrogate': 0, 
      'BiomedicalConcept': 0,
      'BiomedicalConceptProperty': 0,
      'ResponseCode': 0,
      'StudyIntervention': 0,
      'Indication': 0,
      'StudyDesignPopulation': 0,
      'Objective': 0,
      'Endpoint': 0,
      'IntercurrentEvent': 0,
      'Estimand': 0,
      'AnalysisPopulation': 0,
      'StudyDefinitionDocumentVersion': 0,
      'StudyDefinitionDocument': 0,
      'Procedure': 0,
      'TransitionRule': 0,
      'StudyElement': 0,
      'NarrativeContent': 0,
      'NarrativeContentItem': 0,
      'GovernanceDate': 0,
      'GeographicScope': 0,
      'StudyAmendment': 0,
      'StudyAmendmentReason': 0,
      'SubjectEnrollment': 0,
      'SyntaxTemplateDictionary': 0,
      'EligibilityCriterion': 0,
      'Administration': 0,
      'AdministrationDuration': 0,
      'Quantity': 0,
      'Range': 0,
      'StudyCohort': 0,
      'StudyTitle': 0,
      'Masking': 0,
      'ResearchOrganization': 0,
      'StudySite': 0,
      'Condition': 0,
      'ParameterMap': 0,
      'ConditionAssignment': 0,
      'Characteristic': 0,
      'CommentAnnotation': 0,
      'Abbreviation': 0,
      'ReferenceIdentifier': 0,
      'AdministrableProduct': 0,
      'AdministrableProductProperty': 0,
      'AdministrableProductIdentifier': 0,
      'Ingredient': 0,
      'Substance': 0,
      'Strength': 0,
      'AssignedPerson': 0,
      'StudyRole': 0,
      'StudyAmendmentImpact': 0,
      'StudyChange': 0,
      'DocumentContentReference': 0
    }

  def build_id(self, klass):
    klass_name = klass if isinstance(klass, str) else str(klass.__name__)
    self._id_index[klass_name] += 1
    return f"{klass_name}_{self._id_index[klass_name]}"





from usdm_model.identifier import ReferenceIdentifier
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.globals import Globals
from usdm_excel.study_identifiers_sheet.organization import get_organization

class StudyReferencesSheet(BaseSheet):

  SHEET_NAME = 'studyReferences'
  
  def __init__(self, file_path, globals: Globals):
    try:
      self.items = []
      self.organizations = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, optional=True)
      if self.success:
        self.process_sheet()
    except Exception as e:
      self._sheet_exception(e)
      
  def process_sheet(self):
    self.items = []
    for index, row in self.sheet.iterrows():
      organisation = get_organization(self, index)
      if organisation:
        self.organizations.append(organisation)
        text = self.read_cell_by_name(index, 'studyIdentifier')
        type = self.read_cdisc_klass_attribute_cell_by_name('ReferenceIdentifier', 'type', index, ['referenceType'])     
        item = self.create_object(ReferenceIdentifier, {'text': text, 'type': type, 'scopeId': organisation.id})
        if item:
          self.items.append(item)
          self.globals.cross_references.add(item.text, item)         



from usdm_excel.base_manager import BaseManager
from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging

class TemplateManager(BaseManager):
  
  DEFAULT_TEMPLATE_NAME = 'SPONSOR'
  DEFAULT_SHEET_NAME = 'document'

  def __init__(self, errors_and_logging: ErrorsAndLogging):
    super().__init__(errors_and_logging)
    self.add(self.DEFAULT_TEMPLATE_NAME, self.DEFAULT_SHEET_NAME)
    self.default_template = self.DEFAULT_TEMPLATE_NAME
    self.default_sheet = self.DEFAULT_SHEET_NAME

  def tidy(self, sheet_names_present: list):
    new_items = {}
    for template, sheet in self._items.items():
      if sheet in sheet_names_present:
        new_items[template] = sheet
    self._items = new_items

  def templates(self):
    return list(self._items.keys())    



from tests.test_factory import Factory
from usdm_excel.range_type import RangeType

def test_range_type(mocker, globals):
  test_data = [
    ('1..1 Days', "1", "1", '1..1 Days', 'Days', 'C25301'),
    (' -1..1 days', "-1", "1", '-1..1 days', 'days', 'C25301'),
    ('5 .. 10 weeks ', "5", "10", '5 .. 10 weeks', 'weeks', 'C29844'),
  ]
  for index, test in enumerate(test_data):
    item = RangeType(test[0], globals)
    assert(item.lower) == test[1]
    assert(item.upper) == test[2]
    assert(item.label) == test[3]
    assert(item.units) == test[4]
    assert(item.units_code.code) == test[5]
    assert(item.errors) == [] 

def test_range_type_error(mocker, globals):
  test_data = [
    ('1.. Days',"Could not decode the range value, possible typographical errors '1.. Days'"),
    ('-1.1 days',"Could not decode the range value, possible typographical errors '-1.1 days'"),
    ('-1 .. 1',"Could not decode the range value, possible typographical errors '-1 .. 1'"),
    (' .. 1 Weeks',"Could not decode the range value ' .. 1 Weeks'"),
    ('1 .. 10 slugs',"Unable to set the units code for the range '1 .. 10 slugs'")
  ]
  for index, test in enumerate(test_data):
    item = RangeType(test[0], globals)
    assert item.errors == [test[1]]



from usdm_excel.base_sheet import BaseSheet
#from usdm_excel.cross_ref import cross_references
#from usdm_excel.id_manager import id_manager
import traceback
from usdm_model.intercurrent_event import IntercurrentEvent
from usdm_model.analysis_population import AnalysisPopulation
from usdm_model.estimand import Estimand
from usdm_model.study_intervention import StudyIntervention
from usdm_model.endpoint import Endpoint
from usdm_excel.globals import Globals

class StudyDesignEstimandsSheet(BaseSheet):

  SHEET_NAME = 'studyDesignEstimands'
  
  def __init__(self, file_path: str, globals: Globals):
    try:
      self.estimands = []
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME)
      current = None
      current_ice_name = None
      current_ice_description = None
      for index, row in self.sheet.iterrows():
        e_name = self.read_cell_by_name(index, ['name', 'xref'])
        e_summary = self.read_cell_by_name(index, "summaryMeasure")
        ap_description = self.read_cell_by_name(index, 'populationDescription')
        ice_name = self.read_cell_by_name(index, ['intercurrentEventName'])
        ice_description = self.read_cell_by_name(index, ['intercurrentEventDescription', 'description'])
        ice_label = self.read_cell_by_name(index, 'label', must_be_present=False)
        ice_strategy = self.read_cell_by_name(index, "intercurrentEventStrategy")
        treatment_xref = self.read_cell_by_name(index, "treatmentXref")
        endpoint_xref = self.read_cell_by_name(index, "endpointXref")
        if not e_summary == "":
          try:
            ap = AnalysisPopulation(id=self.globals.id_manager.build_id(AnalysisPopulation), name=f"AP_{index+1}", text=ap_description) 
          except Exception as e:
            self._general_error(f"Failed to create AnalysisPopulation object", e)
          else:
            try:
              treatment_id = self._get_treatment(treatment_xref)
              endpoint_id = self._get_endpoint(endpoint_xref)
              params = {'name': e_name,
                        'description': '',
                        'label': e_name,
                        'summaryMeasure': e_summary, 
                        'analysisPopulation': ap, 
                        'interventionId': treatment_id, 
                        'variableOfInterestId': endpoint_id, 
                        'intercurrentEvents': []}
              current = self.create_object(Estimand, params)
              #current = Estimand(id=self.globals.id_manager.build_id(Estimand), summaryMeasure=e_summary, analysisPopulation=ap, interventionId=treatment_id, variableOfInterestId=endpoint_id, intercurrentEvents=[])
            except Exception as e:
              self._general_exception(f"Failed to create Estimand object", e)
            else:
              self.estimands.append(current)  
        if current is not None:
          try:
            ice_name = current_ice_name if ice_name == "" else ice_name
            ice_description = current_ice_description if ice_description == "" else ice_description
            ice = IntercurrentEvent(id=self.globals.id_manager.build_id(IntercurrentEvent), name=ice_name, description=ice_description, label=ice_label, strategy=ice_strategy)
            current_ice_name = ice_name
            current_ice_description = ice_description
          except Exception as e:
            self._general_exception(f"Failed to create IntercurrentEvent object", e)
          else:
            current.intercurrentEvents.append(ice)
        else:
          self._general_error("Failed to add IntercurrentEvent, no Estimand set")

    except Exception as e:
      self._sheet_exception(e)

  def _get_treatment(self, name):
    return self._get_cross_reference(StudyIntervention, name, 'study intervention')

  def _get_endpoint(self, name):
    return self._get_cross_reference(Endpoint, name, 'endpoint')



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .analysis_population import AnalysisPopulation
from .intercurrent_event import IntercurrentEvent
from .comment_annotation import CommentAnnotation

class Estimand(ApiBaseModelWithIdNameLabelAndDesc):
  summaryMeasure: str
  analysisPopulation: AnalysisPopulation
  interventionId: str
  variableOfInterestId: str
  intercurrentEvents: List[IntercurrentEvent]
  notes: List[CommentAnnotation] = []
  instanceType: Literal['Estimand']



from tests.mocks.mock_general import *
from tests.mocks.mock_sheet import *
from tests.mocks.mock_ids import *
from tests.mocks.mock_logging import *
from usdm_excel.study_amendment_sheet.study_amendment_sheet import StudyAmendmentSheet
from usdm_excel.option_manager import Options, EmptyNoneOption

def test_create(mocker, globals):
  globals.id_manager.clear()
  globals.option_manager.set(Options.EMPTY_NONE, EmptyNoneOption.NONE.value)
  sheet_data = {
    'number': ['1', '2'], 
    'summary': ['Added section on unblinding', 'Amended Design'], 
    'primaryReason': ['New Safety Information Available', 'Change In Strategy'],
    'secondaryReasons': ['', 'Investigator/Site Feedback'],
    'enrollment': ['Region: Europe=15, Country: USA=20%', 'Global:31%'],
    'date': ['', '']
  }
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  mock_json = mocker.patch("json.load")
  mock_json.side_effect=[{}, {}, {}]
  item = StudyAmendmentSheet("", globals)
  assert len(item.items) == 2
  assert item.items[0].model_dump() == {
    'id': 'StudyAmendment_1', 
    'number': '1', 
    'summary': 'Added section on unblinding', 
    'primaryReason': {
      'id': 'StudyAmendmentReason_1', 
      'code': {'id': 'Code_1', 'code': 'C99904x4', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'New Safety Information Available', 'instanceType': 'Code'}, 
      'otherReason': None, 
      'instanceType': 'StudyAmendmentReason'
    }, 
    'secondaryReasons': [], 
    'changes': [], 
    'impacts': [], 
    'geographicScopes': [
      {
        'id': 'GeographicScope_3', 
        'type': {'id': 'Code_7', 'code': 'C68846', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Global', 'instanceType': 'Code'}, 
        'code': None, 
        'instanceType': 'GeographicScope'
      }
    ], 
    'enrollments': [
      {
        'id': 'SubjectEnrollment_1', 
        'name': 'XXX', 
        'label': None, 
        'description': None, 
        'quantity': {'id': 'Quantity_1', 'value': 15.0, 'unit': None, 'instanceType': 'Quantity'}, 
        'appliesTo': {
          'id': 'GeographicScope_1', 
          'type': {'id': 'Code_3', 'code': 'C41129', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Region', 'instanceType': 'Code'}, 
          'code': {
            'id': 'AliasCode_1', 
            'standardCode': {'id': 'Code_2', 'code': '150', 'codeSystem': 'ISO 3166 1 alpha3', 'codeSystemVersion': '2020-08', 'decode': 'Europe', 'instanceType': 'Code'}, 
            'standardCodeAliases': [], 
            'instanceType': 'AliasCode'
          }, 
          'instanceType': 'GeographicScope'
        }, 
        'appliesToId': None, 
        'instanceType': 'SubjectEnrollment'
      }, 
      {
        'id': 'SubjectEnrollment_2', 
        'name': 'XXX', 
        'label': None, 
        'description': None, 
        'quantity': {
          'id': 'Quantity_2', 
          'value': 20.0, 
          'unit': {
            'id': 'AliasCode_2', 
            'standardCode': {'id': 'Code_4', 'code': 'C25613', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Percentage', 'instanceType': 'Code'}, 
            'standardCodeAliases': [], 'instanceType': 'AliasCode'
          }, 
          'instanceType': 'Quantity'
        }, 
        'appliesTo': {
          'id': 'GeographicScope_2', 
          'type': {'id': 'Code_6', 'code': 'C25464', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Country', 'instanceType': 'Code'}, 
          'code': {
            'id': 'AliasCode_3', 
            'standardCode': {'id': 'Code_5', 'code': 'DNK', 'codeSystem': 'ISO 3166 1 alpha3', 'codeSystemVersion': '2020-08', 'decode': 'Denmark', 'instanceType': 'Code'}, 
            'standardCodeAliases': [], 
            'instanceType': 'AliasCode'
          }, 
          'instanceType': 'GeographicScope'
        }, 
        'appliesToId': None, 
        'instanceType': 'SubjectEnrollment'
      }
    ], 
    'dateValues': [], 
    'previousId': None, 
    'notes': [], 
    'instanceType': 'StudyAmendment'
  }
  assert item.items[1].model_dump() == {
    'id': 'StudyAmendment_2', 
    'number': '2', 
    'summary': 'Amended Design', 
    'primaryReason': {
      'id': 'StudyAmendmentReason_2', 
      'code': {'id': 'Code_8', 'code': 'C99904x7', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Change In Strategy', 'instanceType': 'Code'}, 
      'otherReason': None, 
      'instanceType': 'StudyAmendmentReason'
    }, 
    'secondaryReasons': [
      {
        'code': {
          'code': 'C99904x10',
          'codeSystem': 'http://www.cdisc.org',
          'codeSystemVersion': '2023-12-15',
          'decode': 'Investigator/Site Feedback',
          'id': 'Code_9',
          'instanceType': 'Code',
        },
        'id': 'StudyAmendmentReason_3',
        'instanceType': 'StudyAmendmentReason',
        'otherReason': None,
      },
    ],
    'changes': [], 
    'impacts': [], 
    'geographicScopes': [
      {
        'id': 'GeographicScope_5', 
        'type': {'id': 'Code_12', 'code': 'C68846', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Global', 'instanceType': 'Code'}, 
        'code': None, 
        'instanceType': 'GeographicScope'
      }
    ], 
    'enrollments': [
      {
        'id': 'SubjectEnrollment_3', 
        'name': 'XXX', 
        'label': None, 
        'description': None, 
        'quantity': {
          'id': 'Quantity_3', 
          'value': 31.0, 
          'unit': {
            'id': 'AliasCode_4', 
            'standardCode': {'id': 'Code_10', 'code': 'C25613', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Percentage', 'instanceType': 'Code'}, 
            'standardCodeAliases': [], 'instanceType': 'AliasCode'
          }, 
          'instanceType': 'Quantity'
        }, 
        'appliesTo': {
          'id': 'GeographicScope_4', 
          'type': {'id': 'Code_11', 'code': 'C68846', 'codeSystem': 'http://www.cdisc.org', 'codeSystemVersion': '2023-12-15', 'decode': 'Global', 'instanceType': 'Code'}, 
          'code': None, 
          'instanceType': 'GeographicScope'
        }, 
        'appliesToId': None, 
        'instanceType': 'SubjectEnrollment'
      }
    ], 
    'dateValues': [], 
    'previousId': 'StudyAmendment_1', 
    'notes': [], 
    'instanceType': 'StudyAmendment'
  }
  
def test_create_empty(mocker, globals):
  sheet_data = {}
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentSheet("", globals)
  assert len(item.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  sheet_data = {
    'number': ['1'], 
    'summary': ['Added section on unblinding'], 
    'primaryReason': ['New Safety Information Available'],
    'enrollment': ['Region: Europe=15, Country: USA=20%'],
    'date': ['']
  }
  mea = mock_error_add(mocker, [None, None, None, None, None, None])
  mock_sheet_present(mocker)
  mock_sheet(mocker, globals, sheet_data)
  item = StudyAmendmentSheet("", globals)
  assert mock_called(mea, 1)
  mock_parameters_correct(mea, [mocker.call('studyAmendments', None, None, "Exception. Error [Failed to detect column(s) 'secondaryReasons' in sheet] while reading sheet 'studyAmendments'. See log for additional details.", 40)])



import inspect
from usdm_excel.errors_and_logging.errors_and_logging import ErrorsAndLogging

class CrossRef():

  class PathError(Exception):
    pass

  def __init__(self, errors_and_logging: ErrorsAndLogging):
    self._errors_and_logging = errors_and_logging
    self._references = {}
    self._identifiers = {}

  def clear(self):
    self._references = {}
    self._identifiers = {}

  def add(self, name, object):
    klass = object.__class__
    try:
      key, id_key = self._key(klass, name, object.id)
      if not key in self._references:
        self._references[key] = object
        self._identifiers[id_key] = object
      else:
        self._error(f"Duplicate cross reference detected, class '{self._klass_name(klass)}' with name '{name}'")
    except Exception as e:
      self._exception(f"Failed to add cross reference detected, class '{self._klass_name(klass)}' with name '{name}'.", e)

  def get(self, klass, name):
    key, id_key = self._key(klass, name, "")
    if key in self._references:
      return self._references[key]
    else:
      self._debug(f"Failed to find by name: klass={self._klass_name(klass)}, name={name}, key={key}:\n\n{'':<9}references={self._references.keys()}")
      return None

  def get_by_id(self, klass, id):
    key, id_key = self._key(klass, "", id)
    if id_key in self._identifiers:
      return self._identifiers[id_key]
    else:
      self._debug(f"Failed to find by id: klass={self._klass_name(klass)}, id={id}, key={id_key}:\n\n{'':<9}identifiers={self._identifiers.keys()}")
      return None

  def get_by_path(self, klass, name, path):
    instance = self.get(klass, name)
    if instance:
      parts = path.split("/")
      attribute = parts[0].replace('@', '')  
      if len(parts) == 1:
        return instance, attribute
      elif len(parts) % 2 == 1:
        for index in range(1,len(parts),2):
          try:
            instance = getattr(instance, attribute)
          except AttributeError as e:
            raise self.PathError(f"Failed to translate reference path '{path}', attribute '{attribute}' was not found")
          attribute = parts[index+1].replace('@', '')
          if not parts[index] == instance.__class__.__name__:
            raise self.PathError(f"Failed to translate reference path '{path}', class mismtach, expecting '{parts[index]}', found '{instance.__class__.__name__}'")
        if instance and attribute:
          if not self.get_by_id(instance.__class__, instance.id):
            self.add(instance.id, instance)
          return instance, attribute
        else:
          raise self.PathError(f"Failed to translate reference path '{path}', path was not found")
      else:
        raise self.PathError(f"Failed to translate reference path '{path}', format error")
    else:
      raise self.PathError(f"Failed to translate reference path '{path}', could not find start instance '{klass}', '{name}'")

  def _key(self, klass, name, id):
    klass_name = self._klass_name(klass)
    return f"{klass_name}.{name}", f"{klass_name}.{id}"

  def _klass_name(self, klass):
    return klass if isinstance(klass, str) else klass.__name__

  def _error(self, message):
    self._errors_and_logging.error(message)

  def _debug(self, message):
    self._errors_and_logging.debug(message)

  def _exception(self, message, e):
    self._errors_and_logging.exception(message, e)



from typing import List, Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity import Quantity
from .administration_duration import AdministrationDuration
from .alias_code import AliasCode
from .comment_annotation import CommentAnnotation

class Administration(ApiBaseModelWithIdNameLabelAndDesc):
  duration:	AdministrationDuration
  dose:	Quantity
  route: AliasCode
  frequency: AliasCode
  administrableProductId: Union[str, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['Administration']



import pytest
import pandas as pd
from usdm_excel.abbreviation_sheet.abbreviation_sheet import AbbreviationSheet
from usdm_model.comment_annotation import CommentAnnotation
from usdm_excel.globals import Globals
from tests.test_factory import Factory

xfail = pytest.mark.xfail

def test_create(mocker, globals):
  data = {
    'abbreviatedText': ['A1', 'A2', 'A3'], 
    'expandedText': ['Annotation Text One', 'Annotation Text Two', 'Annotaiton Text Three']
  }
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Abbreviation_1', 'Abbreviation_2', 'Abbreviation_3']
  _setup(mocker, globals, data)
  item = AbbreviationSheet("", globals)
  assert len(item.items) == 3
  assert item.items[0].id == 'Abbreviation_1'
  assert item.items[0].abbreviatedText == 'A1'
  assert item.items[1].id == 'Abbreviation_2'
  assert item.items[1].abbreviatedText == 'A2'
  assert item.items[1].expandedText == 'Annotation Text Two'
  assert item.items[2].id == 'Abbreviation_3'
  assert item.items[2].abbreviatedText == 'A3'

def test_create_with_notes(factory, mocker, globals):
  data = {
    'abbreviatedText': ['A1'], 
    'expandedText': ['Annotation Text One'],
    'notes': 'N1, N2'
  }
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Note_1', 'Note_2', 'Note_3', 'Abbreviation_1']
  _setup(mocker, globals, data)
  _create_notes(factory, globals)
  item = AbbreviationSheet("", globals)
  assert len(item.items) == 1
  assert item.items[0].id == 'Abbreviation_1'
  assert item.items[0].abbreviatedText == 'A1'
  assert item.items[0].notes[0].text == 'Note 1'
  assert item.items[0].notes[0].codes == []

def test_create_empty(mocker, globals):
  data = {}
  _setup(mocker, globals, data)
  item = AbbreviationSheet("", globals)
  assert len(item.items) == 0

def test_read_cell_by_name_error(mocker, globals):
  data = {
    'abbreviatedText': ['A1'], 
  }
  mock_error = mocker.patch("usdm_excel.errors_and_logging.errors.Errors.add")
  mock_id = mocker.patch("usdm_excel.id_manager.IdManager.build_id")
  mock_id.side_effect=['Code_1', 'Abbreviation_1']
  _setup(mocker, globals, data)
  item = AbbreviationSheet("", globals)
  mock_error.assert_called()
  assert mock_error.call_args[0][0] == "abbreviations"
  assert mock_error.call_args[0][1] == None
  assert mock_error.call_args[0][2] == None
  assert mock_error.call_args[0][3] == "Exception. Failed to create Abbreviation object. See log for additional details."
  
def _setup(mocker, globals, data):
  globals.cross_references.clear()
  mock_present = mocker.patch("usdm_excel.base_sheet.BaseSheet._sheet_present")
  mock_present.side_effect=[True]
  mocked_open = mocker.mock_open(read_data="File")
  mocker.patch("builtins.open", mocked_open)
  mock_read = mocker.patch("pandas.read_excel")
  mock_read.return_value = pd.DataFrame.from_dict(data)

def _create_notes(factory: Factory, globals: Globals):
  notes = {
    'N1': {'text': 'Note 1', 'codes': []},
    'N2': {'text': 'Note 2', 'codes': []},
    'N3': {'text': 'Note 3', 'codes': []},
  }
  for name, note in notes.items():
    item = factory.item(CommentAnnotation, note)
    globals.cross_references.add(name, item)



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity import Quantity
from .administration import Administration
from .code import Code
from .comment_annotation import CommentAnnotation

class StudyIntervention(ApiBaseModelWithIdNameLabelAndDesc):
  role:	Code
  type:	Code
  minimumResponseDuration: Union[Quantity, None] = None
  codes: List[Code] = []
  administrations: List[Administration] = []
  productDesignation:	Code
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyIntervention']



import traceback
from usdm_excel.base_sheet import BaseSheet
from usdm_excel.option_manager import Options, EmptyNoneOption
from usdm_excel.globals import Globals

class ConfigurationSheet(BaseSheet):

  SHEET_NAME = 'configuration'
  
  PARAMS_NAME_COL = 0
  PARAMS_VALUE_COL = 1

  def __init__(self, file_path: str, globals: Globals):
    try:
      super().__init__(file_path=file_path, globals=globals, sheet_name=self.SHEET_NAME, header=None)
      self.globals.option_manager.set(Options.EMPTY_NONE, EmptyNoneOption.NONE)
      self._process_sheet()
      self.globals.template_manager.tidy(self._get_sheet_names(file_path))
      
    except Exception as e:
      self._sheet_exception(e)

  def _process_sheet(self):
    for rindex, row in self.sheet.iterrows():
      raw_name = self.read_cell(rindex, self.PARAMS_NAME_COL)
      name = raw_name.strip().upper()  
      value = self.read_cell(rindex, self.PARAMS_VALUE_COL)
      if name == 'CT VERSION':
        parts = value.split('=')
        if len(parts) == 2:
          self.globals.ct_version_manager.add(parts[0].strip(), parts[1].strip())
        else:
          self._error(rindex, 2, "Badly formatted CT VERSION, should be of the form <CT name> = <version>")
      elif name == 'EMPTY NONE':
        if value.strip().upper() == 'EMPTY':
          self.globals.option_manager.set(Options.EMPTY_NONE, EmptyNoneOption.EMPTY)
      elif name == 'TEMPLATE':
        parts = value.split('=')
        if len(parts) == 2:
          self.globals.template_manager.add(parts[0].strip(), parts[1].strip())
          self._general_info(f"Mapped template '{parts[0].strip()}' to '{parts[1].strip()}'")
        else:
          self._error(rindex, 2, "Badly formatted TEMPLATE, should be of the form <type> = <sheet name>")
      elif name == 'USE TEMPLATE':
        self._general_warning("The USE TEMPLATE option is now deprecated and will be ignored.")
      elif name == 'USDM VERSION':
        self._general_warning("The USDM VERSION option is now deprecated and will be ignored.")
      elif name == 'SDR PREV NEXT':
        self._general_warning("The SDR PREV NEXT option is now deprecated and will be ignored.")
      elif name == 'SDR ROOT':
        self._general_warning("The SDR ROOT option is now deprecated and will be ignored.")
      elif name == 'SDR DESCRIPTION':
        self._general_warning("The SDR DESCRIPTION option is now deprecated and will be ignored.")
      else:
        self._error(rindex, 1, "Unrecognized configuration keyword '{raw_name}")



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithId
from .study_amendment_reason import StudyAmendmentReason
from .study_change import StudyChange
from .study_amendment_impact import StudyAmendmentImpact
from .geographic_scope import GeographicScope
from .subject_enrollment import SubjectEnrollment
from .governance_date import GovernanceDate
from .comment_annotation import CommentAnnotation

class StudyAmendment(ApiBaseModelWithId):
  number: str
  summary: str
  primaryReason: StudyAmendmentReason
  secondaryReasons: List[StudyAmendmentReason] = []
  changes: List[StudyChange] = [] # Not in API
  impacts: List[StudyAmendmentImpact] = []
  geographicScopes: List[GeographicScope]
  enrollments: List[SubjectEnrollment] = []
  dateValues: List[GovernanceDate] = []
  previousId: Union[str, None] = None
  notes: List[CommentAnnotation] = []
  instanceType: Literal['StudyAmendment']
